from pyplanet.apps.config import AppConfig
from pyplanet.apps.contrib.live_rankings.views import LiveRankingsWidget

from pyplanet.apps.core.trackmania import callbacks as tm_signals
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals

from pyplanet.contrib.setting import Setting


class LiveRankings(AppConfig):
	game_dependencies = ['trackmania', 'trackmania_next']
	app_dependencies = ['core.maniaplanet', 'core.trackmania']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.current_rankings = []
		self.points_repartition = []
		self.current_finishes = []
		self.widget = None
		self.dedimania_enabled = False

		self.setting_rankings_amount = Setting(
			'rankings_amount', 'Amount of rankings to display', Setting.CAT_BEHAVIOUR, type=int,
			description='Amount of live rankings to display (minimum: 15).',
			default=15
		)
		self.setting_nadeo_live_ranking = Setting(
			'nadeo_live_ranking', 'Show the Nadeo live rankings widget', Setting.CAT_BEHAVIOUR, type=bool,
			description='Show the Nadeo live rankings widgets besides the live rankings widget.', default=True,
			change_target=self.nadeo_widget_change
		)

	async def on_start(self):
		# Init settings.
		await self.context.setting.register(
			self.setting_rankings_amount, self.setting_nadeo_live_ranking
		)

		# Register signals
		self.context.signals.listen(mp_signals.map.map_start, self.map_start)
		self.context.signals.listen(mp_signals.flow.round_start, self.round_start)
		self.context.signals.listen(tm_signals.finish, self.player_finish)
		self.context.signals.listen(tm_signals.waypoint, self.player_waypoint)
		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)
		self.context.signals.listen(tm_signals.give_up, self.player_giveup)
		self.context.signals.listen(tm_signals.scores, self.scores)
		self.context.signals.listen(tm_signals.warmup_start, self.warmup_start)
		self.context.signals.listen(tm_signals.warmup_end, self.warmup_end)


		# Make sure we move the multilap_info and disable the checkpoint_ranking and round_scores elements.
		if self.instance.game.game in ['tm', 'sm']:
			self.instance.ui_manager.properties.set_visibility('checkpoint_ranking', False)
			self.instance.ui_manager.properties.set_visibility('round_scores', await self.setting_nadeo_live_ranking.get_value())
			self.instance.ui_manager.properties.set_attribute('round_scores', 'pos', '-126.5 80. 150.')
			self.instance.ui_manager.properties.set_attribute('multilap_info', 'pos', '107., 88., 5.')

		self.dedimania_enabled = ('dedimania' in self.instance.apps.apps and 'dedimania' not in self.instance.apps.unloaded_apps)

		self.widget = LiveRankingsWidget(self)
		await self.widget.display()

		scores = None
		try:
			scores = await self.instance.gbx('Trackmania.GetScores')
		except:
			pass

		if scores:
			await self.handle_scores(scores['players'])
			await self.widget.display()

		await self.get_points_repartition()

	async def nadeo_widget_change(self, *args, **kwargs):
		if self.instance.game.game in ['tm', 'sm']:
			self.instance.ui_manager.properties.set_visibility('round_scores', await self.setting_nadeo_live_ranking.get_value())
			await self.instance.ui_manager.properties.send_properties()

	def is_mode_supported(self, mode):
		mode = mode.lower()
		return mode.startswith('timeattack') or mode.startswith('rounds') or mode.startswith('team') or \
			   mode.startswith('laps') or mode.startswith('cup') or mode.startswith('trackmania/tm_timeattack_online') or \
			   mode.startswith('trackmania/tm_rounds_online') or mode.startswith('trackmania/tm_teams_online') or \
			   mode.startswith('trackmania/tm_laps_online') or mode.startswith('trackmania/tm_cup_online')

	async def scores(self, section, players, **kwargs):
		if section == 'PreEndRound':
			# Do not update the live rankings on the 'pre end round'-stage.
			# This will make the points added disappear without updating the actual scores.
			return

		await self.handle_scores(players)
		await self.widget.display()

	async def handle_scores(self, players):
		self.current_rankings = []
		self.current_finishes = []

		current_script = (await self.instance.mode_manager.get_current_script()).lower()
		if 'timeattack' in current_script or 'trackmania/tm_timeattack_online' in current_script:
			for player in players:
				if 'best_race_time' in player:
					if player['best_race_time'] != -1:
						new_ranking = dict(login=player['player'].login, nickname=player['player'].nickname, score=player['best_race_time'])
						self.current_rankings.append(new_ranking)
				elif 'bestracetime' in player:
					if player['bestracetime'] != -1:
						new_ranking = dict(login=player['login'], nickname=player['name'], score=player['bestracetime'])
						self.current_rankings.append(new_ranking)

			self.current_rankings.sort(key=lambda x: x['score'])
		elif 'rounds' in current_script or 'team' in current_script or 'cup' in current_script or \
			'trackmania/tm_rounds_online' in current_script or 'trackmania/tm_teams_online' in current_script or 'trackmania/tm_cup_online' in current_script:
			for player in players:
				if 'map_points' in player:
					if player['map_points'] != -1:
						new_ranking = dict(login=player['player'].login, nickname=player['player'].nickname, score=player['map_points'], points_added=0)
						self.current_rankings.append(new_ranking)
				elif 'mappoints' in player:
					if player['mappoints'] != -1:
						new_ranking = dict(login=player['login'], nickname=player['name'], score=player['mappoints'], points_added=0)
						self.current_rankings.append(new_ranking)

			self.current_rankings.sort(key=lambda x: x['score'])
			self.current_rankings.reverse()

	async def map_start(self, map, restarted, **kwargs):
		self.current_rankings = []
		self.current_finishes = []
		self.dedimania_enabled = ('dedimania' in self.instance.apps.apps and 'dedimania' not in self.instance.apps.unloaded_apps)
		await self.get_points_repartition()
		await self.widget.display()

	async def round_start(self, count, time):
		await self.get_points_repartition()

	async def player_connect(self, player, is_spectator, source, signal):
		await self.widget.display(player=player)

	async def warmup_start(self):
		self.current_rankings = []
		self.current_finishes = []
		await self.get_points_repartition()
		await self.widget.display()

	async def warmup_end(self):
		self.current_rankings = []
		self.current_finishes = []
		await self.get_points_repartition()
		await self.widget.display()

	async def player_giveup(self, time, player, flow):
		current_script = (await self.instance.mode_manager.get_current_script()).lower()
		if 'laps' not in current_script and 'trackmania/tm_laps_online' not in current_script:
			return

		current_rankings = [x for x in self.current_rankings if x['login'] == player.login]
		if len(current_rankings) > 0:
			current_ranking = current_rankings[0]
			current_ranking['giveup'] = True

		await self.widget.display()

	async def player_waypoint(self, player, race_time, flow, raw):
		current_script = (await self.instance.mode_manager.get_current_script()).lower()
		if 'laps' not in current_script and 'trackmania/tm_laps_online' not in current_script:
			return

		current_rankings = [x for x in self.current_rankings if x['login'] == player.login]
		if len(current_rankings) > 0:
			current_ranking = current_rankings[0]
			current_ranking['score'] = raw['racetime']
			current_ranking['cps'] = (raw['checkpointinrace'] + 1)
			current_ranking['best_cps'] = (self.current_rankings[0]['cps'])
			current_ranking['finish'] = raw['isendrace']
			current_ranking['cp_times'] = raw['curracecheckpoints']
			current_ranking['giveup'] = False
		else:
			best_cps = 0
			if len(self.current_rankings) > 0:
				best_cps = (self.current_rankings[0]['cps'])
			new_ranking = dict(login=player.login, nickname=player.nickname, score=raw['racetime'], cps=(raw['checkpointinrace'] + 1), best_cps=best_cps, cp_times=raw['curracecheckpoints'], finish=raw['isendrace'], giveup=False)
			self.current_rankings.append(new_ranking)

		self.current_rankings.sort(key=lambda x: (-x['cps'], x['score']))
		await self.widget.display()

	async def player_finish(self, player, race_time, lap_time, cps, flow, raw, **kwargs):
		current_script = (await self.instance.mode_manager.get_current_script()).lower()
		if 'laps' in current_script or 'trackmania/tm_laps_online' in current_script:
			await self.player_waypoint(player, race_time, flow, raw)
			return

		if 'timeattack' in current_script or 'trackmania/tm_timeattack_online' in current_script:
			current_rankings = [x for x in self.current_rankings if x['login'] == player.login]
			score = lap_time
			if len(current_rankings) > 0:
				current_ranking = current_rankings[0]

				if score < current_ranking['score']:
					current_ranking['score'] = score
					self.current_rankings.sort(key=lambda x: x['score'])
					await self.widget.display()
			else:
				new_ranking = dict(login=player.login, nickname=player.nickname, score=score)
				self.current_rankings.append(new_ranking)
				self.current_rankings.sort(key=lambda x: x['score'])
				await self.widget.display()

			return

		if 'rounds' in current_script or 'team' in current_script or 'cup' in current_script or \
		'trackmania/tm_rounds_online' in current_script or 'trackmania/tm_teams_online' in current_script or 'trackmania/tm_cup_online' in current_script:

			new_finish = dict(login=player.login, nickname=player.nickname, score=race_time)
			self.current_finishes.append(new_finish)

			new_finish_rank = len(self.current_finishes) - 1
			new_finish['points_added'] = self.points_repartition[new_finish_rank] \
				if len(self.points_repartition) > new_finish_rank \
				else self.points_repartition[(len(self.points_repartition) - 1)]

			current_ranking = next((item for item in self.current_rankings if item['login'] == player.login), None)
			if current_ranking is not None:
				current_ranking['points_added'] = new_finish['points_added']
			else:
				new_finish['score'] = 0
				self.current_rankings.append(new_finish)

			self.current_rankings.sort(key=lambda x: (-x['score'], -x['points_added']))
			await self.widget.display()
			return

	async def get_points_repartition(self):
		current_script = (await self.instance.mode_manager.get_current_script()).lower()
		if 'rounds' in current_script or 'team' in current_script or 'cup' in current_script or \
		'trackmania/tm_rounds_online' in current_script or 'trackmania/tm_teams_online' in current_script or 'trackmania/tm_cup_online' in current_script:
			points_repartition = await self.instance.gbx('Trackmania.GetPointsRepartition')
			self.points_repartition = points_repartition['pointsrepartition']
		else:
			# Reset the points repartition array.
			self.points_repartition = []
			self.current_finishes = []



"""
The PIP module contains utilities to upgrade packages in the installation of PyPlanet from the application itself.
"""
import logging
import os
import site
import subprocess
import sys
import tempfile


class Pip:
	def __init__(self):
		self.command = None
		self.virtual_env = None
		self.install_dir = None
		self.pyplanet_dir = None
		self.writable = False
		self.user_flag = False
		self.can_use_user_flag = False

		self.is_supported = False

		self.investigate_environment()

	@classmethod
	def autodetect_pip(cls):
		commands = [[sys.executable, "-m", "pip"],
					[os.path.join(os.path.dirname(sys.executable), "pip.exe" if sys.platform == "win32" else "pip")],

					# this should be our last resort since it might fail thanks to using pip programmatically like
					# that is not officially supported or sanctioned by the pip developers
					[sys.executable, "-c", "import sys; sys.argv = ['pip'] + sys.argv[1:]; import pip; pip.main()"]]

		for command in commands:
			p = subprocess.Popen(command + ['--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
			_, _ = p.communicate()
			if p.returncode == 0:
				logging.getLogger(__name__).info("Using \"{}\" as command to invoke pip".format(" ".join(command)))
				return command

		return None

	def investigate_environment(self):
		"""
		This method investigates the environment of the project (checks for virtualenv/pyenv/--user or system wide installation).
		"""
		self.command = Pip.autodetect_pip()
		if not self.command:
			return
		self.pyplanet_dir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))

		# Checking PIP installation and compatibility...
		test_package = os.path.join(os.path.realpath(os.path.dirname(__file__)), 'pip_test_pkg')
		temp_file = tempfile.NamedTemporaryFile()
		data = dict()

		try:
			p = subprocess.Popen(
				self.command + ['install', '.'],
				stdout=subprocess.PIPE, stderr=subprocess.PIPE,
				cwd=test_package,
				env=dict(
					TESTBALLOON_OUTPUT=temp_file.name
				)
			)
			_, _ = p.communicate()

			with open(temp_file.name) as f:
				for line in f:
					key, value = line.split("=", 2)
					data[key] = value
		except Exception as e:
			logging.getLogger(__name__).error("Failed to check if PIP is supported. Failed to investigate PIP installation.")
			logging.getLogger(__name__).exception(e)
			return

		install_dir_str = data.get("PIP_INSTALL_DIR", None)
		virtual_env_str = data.get("PIP_VIRTUAL_ENV", None)
		writable_str = data.get("PIP_WRITABLE", None)

		if install_dir_str is not None and virtual_env_str is not None and writable_str is not None:
			self.install_dir = install_dir_str.strip()
			self.virtual_env = virtual_env_str.strip() == "True"
			self.writable = writable_str.strip() == "True"

			self.can_use_user_flag = not self.virtual_env and site.ENABLE_USER_SITE

			self.is_supported = self.writable or self.can_use_user_flag
			self.user_flag = not self.writable and self.can_use_user_flag

			logging.getLogger(__name__).info(
				"pip installs to {} (writable -> {}), --user flag needed -> {}, virtual env -> {}".format(
					self.install_dir,
					"yes" if self.writable else "no",
					"yes" if self.user_flag else "no",
					"yes" if self.virtual_env else "no"
				)
			)
			logging.getLogger(__name__).info("==> pip ok -> {}".format("yes" if self.is_supported else "NO!"))
		else:
			logging.getLogger(__name__).error(
				"Could not detect desired output from pip_test_pkg install, got this instead: {!r}".format(data)
			)

	def install(self, package, target_version=None):
		"""
		Install (or upgrade) package to target version given or to latest if no target version given.

		:param package: Package name
		:param target_version: Target version
		:return: Return code from PIP, stdout and stderr
		"""
		if not self.is_supported:
			raise Exception('Pip environment is not supported!')

		command = self.command + ['install']
		if self.user_flag:
			command += ['--user']
		command += ['-U', '{}{}'.format(
			package, '=={}'.format(target_version) if target_version else ''
		)]

		p = subprocess.Popen(
			command,
			stdout=subprocess.PIPE, stderr=subprocess.PIPE,
			cwd=self.pyplanet_dir,
		)
		stdout, stderr = p.communicate()

		return p.returncode, stdout, stderr



"""
This file contains the contrib mode signals, related to the current script/mode.
"""
from pyplanet.core.events import Signal as _Signal, handle_generic
from pyplanet.core.events.manager import SignalManager as _SignalManager


script_mode_changed = _Signal(
	code='script_mode_changed',
	namespace='contrib.mode',
	process_target=handle_generic
)
"""
Is called after a new script has been loaded and became active!. Reporting two parameters:

:param unloaded_script: Old script name.
:param loaded_script: New and just loaded script.
"""

_SignalManager.register_signal([
	script_mode_changed
])



"""
The database migrator class has logic for migrating existing models and holds some information like actual versions
and differences.
"""
import asyncio

import importlib
import logging
import peewee

from playhouse.sqlite_ext import SqliteExtDatabase
from playhouse.migrate import (
	PostgresqlMigrator, SqliteMigrator, MySQLMigrator
)

from pyplanet.core.exceptions import ImproperlyConfigured

logger = logging.getLogger(__name__)


class Migrator:
	def __init__(self, instance, db):
		"""
		Initiate migrator.

		:param instance: Instance of controller
		:param db: Database instance.
		:type instance: pyplanet.core.instance.Instance
		:type db: pyplanet.core.db.database.Database
		"""
		self.instance = instance
		self.db = db
		self.migrator = self.__get_migrator()

		self.pass_migrations = set()

	def __get_migrator(self):
		if isinstance(self.db.engine, peewee.SqliteDatabase) or isinstance(self.db.engine, SqliteExtDatabase):
			return SqliteMigrator(self.db.engine)
		elif isinstance(self.db.engine, peewee.MySQLDatabase):
			return MySQLMigrator(self.db.engine)
		elif isinstance(self.db.engine, peewee.PostgresqlDatabase):
			return PostgresqlMigrator(self.db.engine)
		raise ImproperlyConfigured('Database engine doesn\'t support Migrations!')

	async def create_tables(self):
		creating = list()
		for name, (app, name, model) in self.db.registry.models.items():
			if not model.table_exists():
				creating.append(model)
				self.pass_migrations.add(app.label)

		self.db.engine.create_tables(creating, safe=True)

	async def check(self):
		"""
		Check the database health.
		"""
		try:
			if isinstance(self.migrator, MySQLMigrator):
				cursor = self.db.engine.execute_sql(
					'SELECT DEFAULT_COLLATION_NAME FROM information_schema.SCHEMATA WHERE SCHEMA_NAME LIKE %s;',
					self.db.engine.database
				)
				result = cursor.fetchone()
				if len(result) == 1:
					if result[0] != 'utf8mb4_unicode_ci':
						logger.error(
							'Your database, tables and column collate is \'{}\' and it should be \'utf8mb4_unicode_ci\'! '
							'Please change your database collate right now!'.format(result[0])
						)
						logger.warning(
							'Please read the information on this page on how to convert your collate: '
							'http://www.pypla.net/en/stable/howto/dbcollate.html'
						)
						logger.info('Wait 5 seconds to ignore!... (We strongly advice to change it!)')
						await asyncio.sleep(5)
		except:
			pass  # Totally ignore.

	async def migrate(self):
		"""
		Migrate all models.

		:return:
		"""
		from .models.migration import Migration

		# Create tables + skip migrations.
		await self.create_tables()

		# Look for app migrations that are not yet applied.
		for app, migration_files in self.db.registry.app_migrations.items():
			if not migration_files:
				continue

			# Get module path + current applied migrations.
			app_module = self.instance.apps.apps[app].module.__name__
			applied_migrations = [p.name for p in Migration.select().where(
				(Migration.app == app) & (Migration.applied == True)
			)]

			# For each migration file.
			for full_path, folder, name, ext in migration_files:
				if name in applied_migrations:
					continue
				if app in self.pass_migrations:
					# Fake the migration, we just created all the models for this app. Initial setup.
					Migration.create(
						app=app,
						name=name,
						applied=True
					)
					continue

				# Apply migration..
				self.run_migration(name, app, app_module)

				# Log message.
				logger.info('Successfully executed migration: {}'.format(name))

	def run_migration(self, name, app_label, app_module, save_migration=True):
		"""
		Run + apply migration to the actual database.

		:param name: Name of migration.
		:param app_label: App label.
		:param app_module: App module path.
		:param save_migration: Save migration state?
		"""
		from .models.migration import Migration
		mod = importlib.import_module('{}.migrations.{}'.format(app_module, name))

		try:
			with self.db.allow_sync():
				mod.upgrade(self.migrator)

				if save_migration:
					Migration.create(
						app=app_label,
						name=name,
						applied=True
					)
		except Exception as e:
			logger.warning('Can\'t migrate {}.migrations.{}: {}'.format(app_module, name, str(e)))
			raise



import asyncio
import logging

from pyplanet.core import Controller
from pyplanet.core.events import Callback, Signal, handle_generic
from pyplanet.core.exceptions import SignalGlueStop


async def handle_start_line(source, signal, **kwargs):
	player = await Controller.instance.player_manager.get_player(login=source['login'])
	if not player:
		raise SignalGlueStop()
	royal_mode = await Controller.instance.mode_manager.get_current_script() == 'Trackmania/TM_RoyalTimeAttack_Online'
	flow = player.flow
	flow.start_run()

	# TM 2020 Royal Mode.
	if royal_mode:
		flow.handle_start_line_royal(source['time'])

	return dict(
		player=player, time=source['time'], flow=flow
	)


async def handle_waypoint(source, signal, **kwargs):
	player = await Controller.instance.player_manager.get_player(login=source['login'])
	flow = player.flow
	royal_mode = await Controller.instance.mode_manager.get_current_script() == 'Trackmania/TM_RoyalTimeAttack_Online'

	# Custom waypoint handling for TM 2020 Royal Mode.
	if royal_mode and source['isendlap'] and source['isendrace']:
		handled = flow.handle_waypoint_royal(source['racetime'], source['time'], source['blockid'])

		if handled and not len(flow.royal_block_ids) == 5:
			source['isendlap'] = source['isendrace'] = False

			await finish_royal_section.send_robust(source=dict(
				player=player, section_time=flow.royal_section_time, section_nr=len(flow.royal_block_ids),
				block_id=source['blockid'], flow=flow
			), raw=True)
		elif handled and len(flow.royal_block_ids) == 5:
			source['racetime'] = source['laptime'] = flow.royal_total_time
			source['curlapcheckpoints'] = source['curracecheckpoints'] = flow.royal_times

	if not source['isendlap'] and not source['isendrace']:
		return dict(
			player=player, race_time=source['racetime'], flow=flow, raw=source
		)
	elif source['isendlap'] or source['isendrace']:
		# Check if the time is not zero or bellow, if it is, ignore. See #282.
		if source['racetime'] <= 0 or source['laptime'] <= 0:
			raise SignalGlueStop()

		# End flow and call the other signal.
		flow.reset_run()
		if source['isendlap'] and not source['isendrace']:
			flow.start_run()

		await finish.send_robust(source=dict(
			player=player, race_time=source['racetime'], lap_time=source['laptime'],
			cps=source['curlapcheckpoints'], lap_cps=source['curlapcheckpoints'], race_cps=source['curracecheckpoints'],
			flow=flow, is_end_race=source['isendrace'], is_end_lap=source['isendlap'], raw=source,
		), raw=True)

		# Reset royal when driven 5 finishes.
		if royal_mode and len(flow.royal_block_ids) == 5:
			flow.reset_royal(source['time'])
	else:
		logging.warning('Not isendlap!')
	raise SignalGlueStop()


async def handle_give_up(source, signal, **kwargs):
	player = await Controller.instance.player_manager.get_player(login=source['login'])
	flow = player.flow
	royal_mode = await Controller.instance.mode_manager.get_current_script() == 'Trackmania/TM_RoyalTimeAttack_Online'
	flow.reset_run()

	# TM 2020 Royal mode.
	if royal_mode:
		flow.handle_give_up_royal(source['time'])

	return dict(player=player, flow=player.flow, time=source['time'])


async def handle_respawn(source, signal, **kwargs):
	player = await Controller.instance.player_manager.get_player(login=source['login'])
	return dict(
		player=player, flow=player.flow, race_cp=source['checkpointinrace'], lap_cp=source['checkpointinlap'],
		race_time=source['racetime'], lap_time=source['laptime']
	)


async def handle_scores(source, signal, **kwargs):
	async def get_player_scores(data):
		player = await Controller.instance.player_manager.get_player(login=data['login'])
		return dict(
			player=player,
			best_lap_checkpoints=data['bestlapcheckpoints'],
			match_points=data['matchpoints'], rank=data['rank'], best_lap_time=data['bestlaptime'],
			map_points=data['mappoints'], best_race_time=data['bestracetime'],
			round_points=data['roundpoints'], best_race_checkpoints=data['bestracecheckpoints'],

			# TMNext only
			player_account_id=data['accountid'] if 'accountid' in data else None,

			# ManiaPlanet Only (Not for TMNext)
			stunts_score=data['stuntsscore'] if 'stuntsscore' in data else None,
			best_lap_respawns=data['bestlaprespawns'] if 'bestlaprespawns' in data else None,
			best_race_respawns=data['bestracerespawns'] if 'bestracerespawns' in data else None,

			# New since 2.5.0, so check if exist as we want backward compatibility.
			prevracetime=data['prevracetime'] if 'prevracetime' in data else None,
			prevracerespawns=data['prevracerespawns'] if 'prevracerespawns' in data else None,
			prevracecheckpoints=data['prevracecheckpoints'] if 'prevracecheckpoints' in data else None,
			prevstuntsscore=data['prevstuntsscore'] if 'prevstuntsscore' in data else None,
		)

	def get_team_scores(data):
		return dict(
			map_points=data['mappoints'], name=data['name'], id=data['id'], match_points=data['matchpoints'],
			round_points=data['roundpoints'],
		)

	player_scores = await asyncio.gather(*[
		get_player_scores(d) for d in source['players']
	])
	team_scores = [
		get_team_scores(d) for d in source['teams']
	]
	return dict(
		players=player_scores,
		teams=team_scores,
		winner_team=source['winnerteam'],
		use_teams=source['useteams'],
		winner_player=source['winnerplayer'],
		section=source['section']
	)


async def handle_stunt(source, signal, **kwargs):
	player = await Controller.instance.player_manager.get_player(login=source['login'])
	return dict(
		player=player, race_time=source['racetime'], lap_time=source['laptime'], stunt_score=source['stuntsscore'],
		figure=source['figure'], angle=source['angle'], points=source['points'], combo=source['combo'],
		is_straight=source['isstraight'], is_reverse=source['isreverse'], is_master_jump=source['ismasterjump'],
		factor=source['factor']
	)


start_line = Callback(
	call='Script.Trackmania.Event.StartLine',
	namespace='trackmania',
	code='start_line',
	target=handle_start_line,
)
"""
:Signal:
	Player drives off from the start line.
:Code:
	``trackmania:start_line``
:Description:
	Callback sent when a player starts to race (at the end of the 3,2,1,GO! sequence).
:Original Callback:
	`Script` Trackmania.Event.StartLine

:param time: Server time when callback has been sent.
:param player: Player instance
:param flow: Flow class instance.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
:type flow: pyplanet.apps.core.maniaplanet.models.player.PlayerFlow
"""


start_countdown = Callback(
	call='Script.Trackmania.Event.StartCountdown',
	namespace='trackmania',
	code='start_countdown',
	target=handle_start_line,
)
"""
:Signal:
	Player starts his round, the countdown starts right now.
:Code:
	``trackmania:start_countdown``
:Description:
	Callback sent when a player see the 3,2,1,Go! countdown.
:Original Callback:
	`Script` Trackmania.Event.StartCountdown

:param time: Server time when callback has been sent.
:param player: Player instance
:param flow: Flow class instance.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
:type flow: pyplanet.apps.core.maniaplanet.models.player.PlayerFlow
"""


waypoint = Callback(
	call='Script.Trackmania.Event.WayPoint',
	namespace='trackmania',
	code='waypoint',
	target=handle_waypoint,
)
"""
:Signal:
	Player crosses a checkpoint.
:Code:
	``trackmania:waypoint``
:Description:
	Callback sent when a player crosses a checkpoint.
:Original Callback:
	`Script` Trackmania.Event.WayPoint

player=player, race_time=source['racetime'], flow=flow, raw=source

:param race_time: Total race time in milliseconds.
:param player: Player instance
:param flow: Flow class instance.
:param raw: Raw data, prevent to use this!
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
:type flow: pyplanet.apps.core.maniaplanet.models.player.PlayerFlow

.. note ::

	This signal is not called when the player finishes or passes finish line during laps map.

"""


give_up = Callback(
	call='Script.Trackmania.Event.GiveUp',
	namespace='trackmania',
	code='give_up',
	target=handle_give_up,
)
"""
:Signal:
	Player gives up.
:Code:
	``trackmania:give_up``
:Description:
	Callback sent when a player gives up his current run/round.
:Original Callback:
	`Script` Trackmania.Event.GiveUp

:param time: Server time when callback has been sent.
:param player: Player instance
:param flow: Flow class instance.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
:type flow: pyplanet.apps.core.maniaplanet.models.player.PlayerFlow
"""


respawn = Callback(
	call='Script.Trackmania.Event.Respawn',
	namespace='trackmania',
	code='respawn',
	target=handle_respawn,
)
"""
:Signal:
	Player respawn at cp.
:Code:
	``trackmania:respawn``
:Description:
	Callback sent when a player respawns at the last checkpoint/start.
:Original Callback:
	`Script` Trackmania.Event.Respawn

:param player: Player instance
:param flow: Flow class instance.
:param race_cp: Checkpoint times in current **race**.
:param lap_cp: Checkpoint times in current **lap**.
:param race_time: Total race time in milliseconds.
:param lap_time: Current lap time in milliseconds.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
:type flow: pyplanet.apps.core.maniaplanet.models.player.PlayerFlow
"""


request_respawn = Callback(
	call='Script.Trackmania.Event.OnPlayerRequestRespawn',
	namespace='trackmania',
	code='request_respawn',
	target=handle_generic,
)
"""
:Signal:
	Request respawn by player.
:Code:
	``trackmania:request_respawn``
:Description:
	Callback sent when a player requests a respawns.
:Original Callback:
	`Script` Trackmania.Event.OnPlayerRequestRespawn

:param login: Player login
:param player: Player instance
:param time: Time of event
:type login: str
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
:type time: int
"""


stunt = Callback(
	call='Script.Trackmania.Event.Stunt',
	namespace='trackmania',
	code='stunt',
	target=handle_stunt
)
"""
:Signal:
	Player did a stunt.
:Code:
	``trackmania:stunt``
:Description:
	Callback sent when a player did a stunt.
:Original Callback:
	`Script` Trackmania.Event.Stunt

:param player: Player instance
:param race_time: Total race time in milliseconds.
:param lap_time: Current lap time in milliseconds.
:param stunt_score: Current stunt score.
:param figure: Figure of stunt.
:param angle: Angle of stunt.
:param points: Points got by figure.
:param combo: Combo counter
:param is_straight: Is the jump/stunt straight.
:param is_reverse: Is jump/stunt reversed.
:param is_master_jump: Is master jump.
:param factor: Factor multiplier of points (figure).
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


warmup_start = Callback(
	call='Script.Trackmania.WarmUp.Start',
	namespace='trackmania',
	code='warmup_start',
	target=handle_generic
)
"""
:Signal:
	Warmup Starts
:Code:
	``trackmania:warmup_start``
:Description:
	Callback sent when the warmup starts.
:Original Callback:
	`Script` Trackmania.WarmUp.Start
"""


warmup_end = Callback(
	call='Script.Trackmania.WarmUp.End',
	namespace='trackmania',
	code='warmup_end',
	target=handle_generic
)
"""
:Signal:
	Warmup Ends
:Code:
	``trackmania:warmup_end``
:Description:
	Callback sent when the warmup ends.
:Original Callback:
	`Script` Trackmania.WarmUp.End
"""


warmup_start_round = Callback(
	call='Script.Trackmania.WarmUp.StartRound',
	namespace='trackmania',
	code='warmup_start_round',
	target=handle_generic
)
"""
:Signal:
	Warmup Round Starts.
:Code:
	``trackmania:warmup_start_round``
:Description:
	Callback sent when a warm up round start.
:Original Callback:
	`Script` Trackmania.WarmUp.StartRound

:param current: Current round number.
:param total: Total warm up rounds.
"""


warmup_end_round = Callback(
	call='Script.Trackmania.WarmUp.EndRound',
	namespace='trackmania',
	code='warmup_end_round',
	target=handle_generic
)
"""
:Signal:
	Warmup Round Ends.
:Code:
	``trackmania:warmup_end_round``
:Description:
	Callback sent when a warm up round ends.
:Original Callback:
	`Script` Trackmania.WarmUp.EndRound

:param current: Current round number.
:param total: Total warm up rounds.
"""


scores = Callback(
	call='Script.Trackmania.Scores',
	namespace='trackmania',
	code='scores',
	target=handle_scores
)
"""
:Signal:
	Score callback, called after the map. (Around the podium time).
:Code:
	``trackmania:scores``
:Description:
	Teams and players scores.
:Original Callback:
	`Script` Trackmania.Scores

:param players: Player score payload. Including player instance etc.
:param teams: Team score payload.
:param winner_team: The winning team.
:param use_teams: Use teams.
:param winner_player: The winning player.
:param section: Section, current progress of match. Important to check before you save results!!
:type players: list
:type teams: list
"""


finish = Signal(
	namespace='trackmania',
	code='finish',
)
"""
:Signal:
	Player finishes a lap or the race.
:Code:
	``trackmania:finish``
:Description:
	Player finishes a lap or the complete race. Custom signal!.
:Original Callback:
	*None*

:param player: Player instance.
:param race_time: Time in milliseconds of the complete race.
:param lap_time: Time in milliseconds of the current lap.
:param cps: Deprecated!
:param lap_cps: Current lap checkpoint times.
:param race_cps: Complete race checkpoint times.
:param flow: Flow instance.
:param is_end_race: Is this the finish and end of race.
:param is_end_lap: Is this the finish and end of current lap.
:param raw: Prevent to use this!
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
:type flow: pyplanet.apps.core.maniaplanet.models.player.PlayerFlow
:type race_time: int
:type lap_time: int
:type lap_cps: list
:type race_cps: list
:type is_end_race: bool
:type is_end_lap: bool
"""


finish_royal_section = Signal(
	namespace='trackmania',
	code='finish_royal_section',
)
"""
:Signal:
	Player finishes a section of a royal map. (it will also fire waypoint!)
:Code:
	``trackmania:finish_royal_section``
:Description:
	Player finishes a section of a royal map. Custom signal!.
:Original Callback:
	*None*

:param player: Player instance.
:param section_time: Time in milliseconds of the complete race.
:param section_nr: The section number.
:param block_id: Block ID.
:param flow: Flow instance.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
:type flow: pyplanet.apps.core.maniaplanet.models.player.PlayerFlow
:type section_time: int
:type section_nr: int
:type block_id: any
"""


warmup_status = Callback(
	call='Script.Trackmania.WarmUp.Status',
	namespace='trackmania',
	code='warmup_status',
	target=handle_generic
)
"""
:Signal:
	Status of Trackmania warmup. (mostly as response).
:Code:
	``trackmania:warmup_status``
:Description:
	The status of Trackmania's the warmup.
:Original Callback:
	`Script` Trackmania.WarmUp.Status

:param responseid: Internally used. Ignore
:param available: Is warmup available in the game mode. (Boolean).
:param active: Is warmup active and ongoing right now.
:type available: bool
:type active: bool
"""


tmnext_properties = Callback(
	call='Script.Common.UIModules.Properties',
	namespace='trackmania',
	code='common_uimodules_properties',
	target=handle_generic
)
"""
:Signal:
	Common.UIModules.Properties for TM2020
:Code:
	``trackmania:common_uimodules_properties``
:Description:
	Returns UIModules Properties like for instance: Race_Chrono see: ``[{'id': 'Race_Chrono', 'position': [0, -80], 'scale': 1, 'visible': True},``
:Original Callback:
	`Script` Common.UIModules.Properties

:param responseid: Internally used. Ignore
:param uimodules: Collection of all UIModules in TM2020.
:type id Name of the UIModule
:type position Position to place/get Position [0,-80] as example
:type scale Scale measurement
:type visible Boolean True or False
"""


tmnext_ko_elimination = Callback(
	call='Script.Trackmania.Knockout.Elimination',
	namespace='trackmania',
	code='tmnext_ko_elimination',
	target=handle_generic
)
"""
:Signal:
	Knockout Elimination Callback.
:Code:
	``trackmania:knockout_Elimination``
:Description:
	Returns a webserviceId on Player Knockout.
:Original Callback:
	`Script` Trackmania.Knockout.Elimination

:param responseid: Internally used. Ignore
:param account-id: example: ab6fa572-48d7-4b16-a2d7-5c760a05f97b.

"""



import asyncio
import datetime
import logging

from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.apps.contrib.voting.views import VoteWidget
from pyplanet.apps.contrib.voting.vote import Vote
from pyplanet.contrib.command import Command
from pyplanet.contrib.map.exceptions import ModeIncompatible
from pyplanet.contrib.setting import Setting

logger = logging.getLogger(__name__)


class Voting(AppConfig):
	"""
	Chat-based voting plugin.
	"""

	name = 'pyplanet.apps.contrib.voting'
	game_dependencies = ['trackmania', 'trackmania_next', 'shootmania']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		"""
		Initializes the voting plugin.
		"""
		super().__init__(*args, **kwargs)

		self.current_vote = None
		self.widget = None
		self.podium_stage = False
		self.extend_current_count = 0

		self.setting_voting_enabled = Setting(
			'voting_enabled', 'Voting enabled', Setting.CAT_BEHAVIOUR, type=bool,
			description='Whether or not chat-based voting is enabled.',
			default=True
		)

		self.setting_voting_ratio = Setting(
			'voting_ratio', 'Voting ratio', Setting.CAT_BEHAVIOUR, type=float,
			description='Ratio of players that have to vote in favour for the vote to go through (between 0 and 1).',
			default=0.6
		)

		self.setting_voting_timeout = Setting(
			'voting_timeout', 'Voting timeout', Setting.CAT_BEHAVIOUR, type=int,
			description='Timeout in seconds before a vote is cancelled if the required votes aren\'t achieved (0 = no timeout).',
			default=120
		)

		self.setting_callvoting_disable = Setting(
			'callvoting_disable', 'Disable callvoting', Setting.CAT_BEHAVIOUR, type=bool,
			description='Disable callvoting when chat-based voting is enabled.',
			default=True
		)

		self.setting_callvoting_timeout = Setting(
			'callvoting_timeout', 'Callvote time-out (chat disabled)', Setting.CAT_BEHAVIOUR, type=int,
			description='Callvote time-out in milliseconds, set when the chat-voting plugin is disabled (0 = disabled).',
			default=60000
		)

		self.setting_enabled_replay = Setting(
			'enabled_replay', 'Replay vote enabled', Setting.CAT_BEHAVIOUR, type=bool,
			description='Whether or not the replay vote is enabled.',
			default=True
		)

		self.setting_enabled_restart = Setting(
			'enabled_restart', 'Restart vote enabled', Setting.CAT_BEHAVIOUR, type=bool,
			description='Whether or not the restart vote is enabled.',
			default=True
		)

		self.setting_enabled_skip = Setting(
			'enabled_skip', 'Skip vote enabled', Setting.CAT_BEHAVIOUR, type=bool,
			description='Whether or not the skip vote is enabled.',
			default=True
		)

		self.setting_enabled_time_extend = Setting(
			'enabled_time_extend', 'Time Extend vote enabled', Setting.CAT_BEHAVIOUR, type=bool,
			description='Whether or not the time extend vote is enabled (only works in TimeAttack mode).',
			default=True
		)

		self.setting_remind_interval = Setting(
			'remind_interval', 'Vote reminder interval', Setting.CAT_BEHAVIOUR, type=int,
			description='Interval in seconds before players are reminded to vote.',
			default=30
		)

		self.setting_extend_max_amount = Setting(
			'extend_max_amount', 'Max number of extends', Setting.CAT_BEHAVIOUR, type=int,
			description='Set max number of extends allowed on map. use -1 for disable.',
			default=-1
		)

	async def on_start(self):
		"""
		Called on starting the application.
		Will register the voting settings, permissions, signals and commands.
		Also creates the voting widget.

		Disables callvoting (server-based voting).
		"""

		await self.context.setting.register(
			self.setting_voting_enabled, self.setting_voting_ratio, self.setting_voting_timeout,
			self.setting_remind_interval, self.setting_enabled_replay, self.setting_enabled_restart,
			self.setting_enabled_skip, self.setting_callvoting_disable, self.setting_callvoting_timeout,
			self.setting_enabled_time_extend, self.setting_extend_max_amount
		)

		await self.instance.permission_manager.register('cancel', 'Cancel the current vote', app=self, min_level=1)
		await self.instance.permission_manager.register('pass', 'Pass the current vote', app=self, min_level=1)

		await self.instance.command_manager.register(
			Command(command='cancel', target=self.cancel_vote, perms='voting:cancel', admin=True,
					description='Cancels the current chat-based vote.'),
			Command(command='pass', target=self.pass_vote, perms='voting:pass', admin=True,
					description='Passes the current chat-based vote.'),
			Command(command='y', aliases=['yes'], target=self.vote_yes,
					description='Votes yes on the current chat-based vote.'),
			Command(command='n', aliases=['no'], target=self.vote_no,
					description='Votes no on the current chat-based vote.'),
			Command(command='replay', target=self.vote_replay,
					description='Starts a chat-based vote to replay the current map.'),
			Command(command='restart', aliases=['res'], target=self.vote_restart,
					description='Starts a chat-based vote to restart the current map.'),
			Command(command='skip', target=self.vote_skip,
					description='Starts a chat-based vote to skip the current map.'),
			Command(command='extend', target=self.vote_extend,
					description='Starts a chat-based vote to extend the playing time on the current map.'),
		)

		self.widget = VoteWidget(self)

		# Register callback.
		self.context.signals.listen(mp_signals.flow.podium_start, self.podium_start)
		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)
		self.context.signals.listen(mp_signals.map.map_start, self.map_start)

		if await self.setting_callvoting_disable.get_value() is True:
			# Disable callvoting
			await self.instance.gbx('SetCallVoteTimeOut', 0)

	async def on_stop(self):
		"""
		Called on unloading the application.
		Enables callvoting (server-based voting).
		"""

		timeout = await self.setting_callvoting_timeout.get_value()
		await self.instance.gbx('SetCallVoteTimeOut', timeout)

	async def player_connect(self, player, is_spectator, source, signal):
		"""
		Called on a player connecting to the server.
		Will display the voting widget to the player if a vote is currently running.

		:param player: player that is joining the server
		:param is_spectator: whether the joining player is a spectator
		"""

		if self.widget is None:
			self.widget = VoteWidget(self)

		# If there is currently a vote, display the voting widget.
		if self.current_vote is not None:
			await self.widget.display(player=player)

	async def podium_start(self, **kwargs):
		"""
		Called when the server switches to the podium (when the map is finished).
		Will cancel the current vote is there is one running.
		Also prohibits starting a new vote.
		"""

		self.podium_stage = True
		self.extend_current_count = 0

		if self.current_vote is not None:
			message = '$0cfVote to $fff{}$0cf has been cancelled.'.format(self.current_vote.action)
			await self.instance.chat(message)

			# Hide the voting widget and reset the current vote
			await self.reset_vote()

	async def map_start(self, *args, **kwargs):
		"""
		Called when a new map is being started.
		Enables starting a new vote again.
		"""

		self.podium_stage = False
		self.extend_current_count = 0

	async def cancel_vote(self, player, data, **kwargs):
		"""
		Admin command: //cancel
		Cancels the current vote if there is one running.

		:param player: player (admin) cancelling the vote
		"""

		if self.current_vote is None:
			message = '$i$f00There is currently no vote in progress.'
			await self.instance.chat(message, player)
			return

		message = '$0cfAdmin $fff{}$z$s$0cf cancelled the current vote to $fff{}$z$s$0cf.'.format(player.nickname,
																								  self.current_vote.action)
		await self.instance.chat(message)

		# Hide the voting widget and reset the current vote
		await self.reset_vote()

	async def pass_vote(self, player, data, **kwargs):
		"""
		Admin command: //pass
		Passes the current vote if there is one running.

		:param player: player (admin) passing the vote
		"""

		if self.current_vote is None:
			message = '$i$f00There is currently no vote in progress.'
			await self.instance.chat(message, player)
			return

		message = '$0cfAdmin $fff{}$z$s$0cf passed the current vote to $fff{}$z$s$0cf.'.format(player.nickname,
																							   self.current_vote.action)
		await self.instance.chat(message)

		await self.current_vote.fire_passed_event(True)

		# Hide the voting widget and reset the current vote
		await self.reset_vote()

	async def vote_added(self, vote, player):
		"""
		Called when a player voted in favour of the current vote.

		:param vote: current vote
		:param player: player voting for the vote
		"""

		if len(vote.votes_current) >= vote.votes_required:
			message = '$fff{}$z$s$0cf voted to $fff{}$0cf.'.format(player.nickname, vote.action)
			await self.instance.chat(message)
		else:
			required_votes = (vote.votes_required - len(vote.votes_current))
			message = '$fff{}$z$s$0cf voted to $fff{}$0cf, $fff{}$0cf more {} needed (use $fffF5$0cf to vote){}.'.format(
				player.nickname, vote.action, required_votes, ('votes' if required_votes > 1 else 'vote'),
				(' ($fff{}$0cf seconds left)'.format(int(round((
																	   self.current_vote.time_limit - datetime.datetime.now()).total_seconds()))) if self.current_vote.time_limit is not None else '')
			)
			await self.instance.chat(message)

	async def vote_removed(self, vote, player):
		"""
		Called when a player voted against the current vote.

		:param vote: current vote
		:param player: player voting against the vote
		"""

		required_votes = (vote.votes_required - len(vote.votes_current))
		message = '$0cfThere {} $fff{}$0cf more {} needed to $fff{}$0cf (use $fffF5$0cf to vote){}.'.format(
			('are' if required_votes > 1 else 'is'), required_votes, ('votes' if required_votes > 1 else 'vote'),
			vote.action,
			(' ($fff{}$0cf seconds left)'.format(int(round((
																   self.current_vote.time_limit - datetime.datetime.now()).total_seconds()))) if self.current_vote.time_limit is not None else '')
		)
		await self.instance.chat(message)

	async def vote_yes(self, player, data, **kwargs):
		"""
		Chat command: /yes
		Called when a player votes in favour of the vote via the chat.

		:param player: player voting
		"""

		if self.current_vote is None:
			message = '$i$f00There is currently no vote in progress.'
			await self.instance.chat(message, player)
			return

		if player.flow.is_spectator:
			message = '$i$f00Only players are allowed to vote.'
			await self.instance.chat(message, player)
			return

		if not await self.current_vote.add_vote(player):
			message = '$i$f00You have already voted on this vote!'
			await self.instance.chat(message, player)
		else:
			await self.widget.hide(player_logins=[player.login])

	async def vote_no(self, player, data, **kwargs):
		"""
		Chat command: /no
		Called when a player votes against the vote via the chat.

		:param player: player voting
		"""

		if self.current_vote is None:
			message = '$i$f00There is currently no vote in progress.'
			await self.instance.chat(message, player)
			return

		if player.flow.is_spectator:
			message = '$i$f00Only players are allowed to vote.'
			await self.instance.chat(message, player)
			return

		await self.current_vote.remove_vote(player)
		message = '$0cfYou have successfully voted $fffno$0cf.'
		await self.instance.chat(message, player)

		await self.widget.hide(player_logins=[player.login])

	async def vote_replay(self, player, data, **kwargs):
		"""
		Chat command: /replay
		Called when a player requests to start a replay vote.

		:param player: player requesting the vote
		"""

		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return

		if self.current_vote is not None:
			message = '$i$f00You cannot start a vote while one is already in progress.'
			await self.instance.chat(message, player)
			return

		if await self.setting_voting_enabled.get_value() is False:
			message = '$i$f00Chat-based voting has been disabled via the server settings!'
			await self.instance.chat(message, player)
			return

		if await self.setting_enabled_replay.get_value() is False:
			message = '$i$f00Replay voting has been disabled via the server settings! Try /extend for time-extension of TA-mode'
			await self.instance.chat(message, player)
			return

		if self.podium_stage:
			message = '$i$f00You cannot start a vote during the podium!'
			await self.instance.chat(message, player)
			return

		if player.flow.is_spectator:
			message = '$i$f00Only players are allowed to vote.'
			await self.instance.chat(message, player)
			return

		if 'jukebox' not in self.instance.apps.apps:
			message = '$i$f00The jukebox plugin is required to start a vote to replay this map.'
			await self.instance.chat(message, player)
			return

		if len(self.instance.apps.apps['jukebox'].jukebox) > 0:
			first_juked = self.instance.apps.apps['jukebox'].jukebox[0]
			if first_juked['map'].uid == self.instance.map_manager.current_map.uid:
				message = '$i$f00This map is already being replayed.'
				await self.instance.chat(message, player)
				return

		await self.create_vote('replay this map', player, self.vote_replay_passed)

		if 'timeattack' in (await self.instance.mode_manager.get_current_script()).lower() \
			and await self.setting_enabled_time_extend.get_value():
			message = '$i$FD4Did you know that you could also vote for extending the time limit with /extend?'
			await self.instance.chat(message, player)

		message = '$fff{}$z$s$0cf wants to $fff{}$0cf, $fff{}$0cf more {} needed (use $fffF5$0cf to vote){}.'.format(
			player.nickname, self.current_vote.action, self.current_vote.votes_required,
			('votes' if self.current_vote.votes_required > 1 else 'vote'),
			(' ($fff{}$0cf seconds left)'.format(int(round((
																   self.current_vote.time_limit - datetime.datetime.now()).total_seconds()))) if self.current_vote.time_limit is not None else '')
		)
		await self.instance.chat(message)

		await self.current_vote.add_vote(player)

	async def vote_replay_passed(self, vote, forced):
		"""
		Called when a replay vote gets passed.

		:param vote: vote that passed
		:param forced: whether the vote was forced passed by an admin
		"""

		if 'jukebox' not in self.instance.apps.apps:
			return

		maps_in_jukebox = [j['map'] for j in self.instance.apps.apps['jukebox'].jukebox]
		if self.instance.map_manager.current_map.uid in [m.uid for m in maps_in_jukebox]:
			drop_map = next((item for item in self.instance.apps.apps['jukebox'].jukebox if
							 item['map'].uid == self.instance.map_manager.current_map.uid), None)
			if drop_map is not None:
				self.instance.apps.apps['jukebox'].jukebox.remove(drop_map)

		self.instance.apps.apps['jukebox'].jukebox = [{'player': vote.requester,
													   'map': self.instance.map_manager.current_map}] + \
													 self.instance.apps.apps['jukebox'].jukebox

		if not forced:
			message = '$0cfVote to $fff{}$0cf has passed.'.format(vote.action)
			await self.instance.chat(message)

		# Hide the voting widget and reset the current vote
		await self.reset_vote()

	async def vote_restart(self, player, data, **kwargs):
		"""
		Chat command: /restart
		Called when a player requests to start a restart vote.

		:param player: player requesting the vote
		"""

		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return

		if self.current_vote is not None:
			message = '$i$f00You cannot start a vote while one is already in progress.'
			await self.instance.chat(message, player)
			return

		if await self.setting_voting_enabled.get_value() is False:
			message = '$i$f00Chat-based voting has been disabled via the server settings!'
			await self.instance.chat(message, player)
			return

		if await self.setting_enabled_restart.get_value() is False:
			message = '$i$f00Restart voting has been disabled via the server settings! Try /extend for time-extension of TA-mode'
			await self.instance.chat(message, player)
			return

		if self.podium_stage:
			message = '$i$f00You cannot start a vote during the podium!'
			await self.instance.chat(message, player)
			return

		if player.flow.is_spectator:
			message = '$i$f00Only players are allowed to vote.'
			await self.instance.chat(message, player)
			return

		await self.create_vote('restart this map', player, self.vote_restart_passed)

		message = '$fff{}$z$s$0cf wants to $fff{}$0cf, $fff{}$0cf more {} needed (use $fffF5$0cf to vote){}.'.format(
			player.nickname, self.current_vote.action, self.current_vote.votes_required,
			('votes' if self.current_vote.votes_required > 1 else 'vote'),
			(' ($fff{}$0cf seconds left)'.format(int(round((
																   self.current_vote.time_limit - datetime.datetime.now()).total_seconds()))) if self.current_vote.time_limit is not None else '')
		)
		await self.instance.chat(message)

		await self.current_vote.add_vote(player)

	async def vote_restart_passed(self, vote, forced):
		"""
		Called when a restart vote gets passed.

		:param vote: vote that passed
		:param forced: whether the vote was forced passed by an admin
		"""

		message = '$0cfVote to $fff{}$0cf has passed.'.format(vote.action)

		# Hide the voting widget and reset the current vote
		await self.reset_vote()

		# Dedimania save vreplay/ghost replays first.
		if 'dedimania' in self.instance.apps.apps:
			logger.info('Saving dedimania (v)replays first!..')
			if hasattr(self.instance.apps.apps['dedimania'], 'podium_start'):
				try:
					await self.instance.apps.apps['dedimania'].podium_start()
				except Exception as e:
					logger.exception(e)

		await self.instance.gbx('RestartMap')

		if not forced:
			await self.instance.chat(message)

	async def vote_skip(self, player, data, **kwargs):
		"""
		Chat command: /skip
		Called when a player requests to start a skip vote.

		:param player: player requesting the vote
		"""

		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return

		if self.current_vote is not None:
			message = '$i$f00You cannot start a vote while one is already in progress.'
			await self.instance.chat(message, player)
			return

		if await self.setting_voting_enabled.get_value() is False:
			message = '$i$f00Chat-based voting has been disabled via the server settings!'
			await self.instance.chat(message, player)
			return

		if await self.setting_enabled_skip.get_value() is False:
			message = '$i$f00Skip voting has been disabled via the server settings!'
			await self.instance.chat(message, player)
			return

		if self.podium_stage:
			message = '$i$f00You cannot start a vote during the podium!'
			await self.instance.chat(message, player)
			return

		if player.flow.is_spectator:
			message = '$i$f00Only players are allowed to vote.'
			await self.instance.chat(message, player)
			return

		await self.create_vote('skip this map', player, self.vote_skip_passed)

		message = '$fff{}$z$s$0cf wants to $fff{}$0cf, $fff{}$0cf more {} needed (use $fffF5$0cf to vote){}.'.format(
			player.nickname, self.current_vote.action, self.current_vote.votes_required,
			('votes' if self.current_vote.votes_required > 1 else 'vote'),
			(' ($fff{}$0cf seconds left)'.format(int(round((
																   self.current_vote.time_limit - datetime.datetime.now()).total_seconds()))) if self.current_vote.time_limit is not None else '')
		)
		await self.instance.chat(message)

		await self.current_vote.add_vote(player)

	async def vote_skip_passed(self, vote, forced):
		"""
		Called when a skip vote gets passed.

		:param vote: vote that passed
		:param forced: whether the vote was forced passed by an admin
		"""

		message = '$0cfVote to $fff{}$0cf has passed.'.format(vote.action)

		# Hide the voting widget and reset the current vote
		await self.reset_vote()

		await self.instance.gbx('NextMap')

		if not forced:
			await self.instance.chat(message)

	async def vote_extend(self, player, data, **kwargs):
		"""
		Chat command: /extend
		Called when a player requests to start an extend vote.

		:param player: player requesting the vote
		"""

		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return

		extend_max = await self.setting_extend_max_amount.get_value()
		if 0 < extend_max <= self.extend_current_count:
			message = f'$i$f00Map has reached extend limit (max {extend_max})'
			await self.instance.chat(message, player)
			return

		if self.current_vote is not None:
			message = '$i$f00You cannot start a vote while one is already in progress.'
			await self.instance.chat(message, player)
			return

		if await self.setting_voting_enabled.get_value() is False:
			message = '$i$f00Chat-based voting has been disabled via the server settings!'
			await self.instance.chat(message, player)
			return

		if await self.setting_enabled_time_extend.get_value() is False:
			message = '$i$f00Skip voting has been disabled via the server settings!'
			await self.instance.chat(message, player)
			return

		if 'timeattack' not in (await self.instance.mode_manager.get_current_script()).lower():
			message = '$i$f00Time Extend voting is only supported in Time Attack modes!'
			await self.instance.chat(message, player)
			return

		if self.podium_stage:
			message = '$i$f00You cannot start a vote during the podium!'
			await self.instance.chat(message, player)
			return

		if player.flow.is_spectator:
			message = '$i$f00Only players are allowed to vote.'
			await self.instance.chat(message, player)
			return

		await self.create_vote('extend time limit on this map', player, self.vote_extend_passed)

		message = '$fff{}$z$s$0cf wants to $fff{}$0cf, $fff{}$0cf more {} needed (use $fffF5$0cf to vote){}.'.format(
			player.nickname, self.current_vote.action, self.current_vote.votes_required,
			('votes' if self.current_vote.votes_required > 1 else 'vote'),
			(' ($fff{}$0cf seconds left)'.format(int(round((
																   self.current_vote.time_limit - datetime.datetime.now()).total_seconds()))) if self.current_vote.time_limit is not None else '')
		)
		await self.instance.chat(message)
		await self.current_vote.add_vote(player)

	async def vote_extend_passed(self, vote, forced):
		"""
		Called when an extend vote gets passed.

		:param vote: vote that passed
		:param forced: whether the vote was forced passed by an admin
		"""

		message = '$0cfVote to $fff{}$0cf has passed.'.format(vote.action)
		self.extend_current_count += 1
		# Hide the voting widget and reset the current vote
		await self.reset_vote()

		try:
			await self.instance.map_manager.extend_ta()
			if not forced:
				await self.instance.chat(message)
		except ModeIncompatible:
			await self.instance.chat('$0cfVote to $fff{}$0cf has failed, current mode not Time Attack?')

	async def vote_reminder(self, vote):
		"""
		Called in a loop to keep informing players via the chat that a vote is open.

		:param vote: vote that is currently open
		"""

		await asyncio.sleep(await self.setting_remind_interval.get_value())

		if self.current_vote is not None:
			required_votes = (self.current_vote.votes_required - len(self.current_vote.votes_current))
			current_required_votes = (vote.votes_required - len(vote.votes_current))
			if self.current_vote.action == vote.action and current_required_votes == required_votes:
				message = '$0cfThere are $fff{}$0cf more {} needed to $fff{}$0cf (use $fffF5$0cf to vote){}.'.format(
					current_required_votes, ('votes' if current_required_votes > 1 else 'vote'),
					self.current_vote.action,
					(' ($fff{}$0cf seconds left)'.format(int(round((
																		   self.current_vote.time_limit - datetime.datetime.now()).total_seconds()))) if self.current_vote.time_limit is not None else '')
				)
				await self.instance.chat(message)

				asyncio.ensure_future(self.vote_reminder(vote))

	async def vote_canceller(self, vote):
		"""
		Called in a loop to keep checking whether the current vote should be cancelled because the time limit is passed.

		:param vote: vote that is currently open
		"""

		await asyncio.sleep(1)

		if self.current_vote is not None and self.current_vote.time_limit is not None:
			if datetime.datetime.now() > self.current_vote.time_limit:
				message = '$0cfVote to $fff{}$0cf has timed out after $fff{}$0cf seconds.'.format(
					self.current_vote.action, await self.setting_voting_timeout.get_value()
				)
				await self.instance.chat(message)

				# Hide the voting widget and reset the current vote
				await self.reset_vote()
			else:
				asyncio.ensure_future(self.vote_canceller(vote))

	async def create_vote(self, action, player, passed_event):
		"""
		Called to create a new current vote.

		:param action: kind of vote that should be created
		:param player: player requesting the vote
		:param passed_event: method to be called when the vote is passed
		"""

		ratio = await self.setting_voting_ratio.get_value()
		players = self.instance.player_manager.count_players

		new_vote = Vote()
		new_vote.action = action
		new_vote.requester = player
		new_vote.votes_current = []
		needed_votes = round(players * ratio)
		if needed_votes > players:
			needed_votes = players
		new_vote.votes_required = needed_votes
		new_vote.vote_added = self.vote_added
		new_vote.vote_removed = self.vote_removed
		new_vote.vote_passed = passed_event

		time_limit = await self.setting_voting_timeout.get_value()
		if time_limit > 0:
			new_vote.time_limit = datetime.datetime.now() + datetime.timedelta(0, time_limit)
			asyncio.ensure_future(self.vote_canceller(new_vote))

		asyncio.ensure_future(self.vote_reminder(new_vote))

		# Set the current vote here, so the widget can access it.
		self.current_vote = new_vote

		# Display the voting widget.
		await self.widget.display()

		# Hide the widget for the vote requester, as that person always votes.
		await self.widget.hide(player_logins=[player.login])

		return new_vote

	async def reset_vote(self):
		"""
		Called to reset the current vote - will hide the widget and reset the variable.
		"""

		await self.widget.hide()
		self.current_vote = None



import asynctest

from pyplanet.core import Controller


class TestCommands(asynctest.TestCase):
	async def target(self, *args, **kwargs):
		self.target_called += 1

	async def num_target(self, player, data, **kwargs):
		self.target_called += len(data.numbers)

	async def test_registering(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()
		await instance.db.initiate()

		self.target_called = 0

		from pyplanet.apps.core.maniaplanet.models import Player
		from pyplanet.contrib.command import Command

		test1 = Command(
			'test',
			self.target,
			aliases=['tst'],
			admin=False,
		)
		await instance.command_manager.register(
			test1
		)

		# Try to fetch the command.
		player, _ = await Player.get_or_create(login='sample-1', nickname='sample-1', level=Player.LEVEL_MASTER)
		await instance.command_manager._on_chat(player, '/test', True)
		await instance.command_manager._on_chat(player, '/tst', True)

		assert self.target_called == 2

	async def test_params(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()
		await instance.db.initiate()

		self.target_called = 0

		from pyplanet.apps.core.maniaplanet.models import Player
		from pyplanet.contrib.command import Command

		test1 = Command(
			'test',
			self.target,
			aliases=['tst'],
			admin=False,
		).add_param('test', type=str, required=True)
		test2 = Command(
			'test2',
			self.num_target,
			admin=True,
		).add_param('numbers', type=int, nargs='*', required=True)
		await instance.command_manager.register(
			test1, test2
		)

		# Try to fetch the command.
		player, _ = await Player.get_or_create(login='sample-1', nickname='sample-1', level=Player.LEVEL_MASTER)
		await instance.command_manager._on_chat(player, '/test ok', True)
		await instance.command_manager._on_chat(player, '/tst okok', True)

		assert self.target_called == 2
		self.target_called = 0

		await instance.command_manager._on_chat(player, '//test2 1 2 3 4 5', True)
		await instance.command_manager._on_chat(player, '/admin test2 4 5', True)

		assert self.target_called == 7



from pyplanet.utils import times


def test_time_parsing():
	raw = 14004
	expect = '0:14.004'
	assert times.format_time(raw) == expect

	raw = 20135
	expect = '0:20.135'
	assert times.format_time(raw) == expect

	raw = 65195
	expect = '1:05.195'
	assert times.format_time(raw) == expect

	raw = 605195
	expect = '10:05.195'
	assert times.format_time(raw) == expect

	raw = 6005195
	expect = '01:40:05.195'
	assert times.format_time(raw) == expect



from peewee import *
from playhouse.migrate import migrate, SchemaMigrator

from ..models import TestModel


def upgrade(migrator: SchemaMigrator):
	migrate(
		migrator.drop_column(TestModel._meta.db_table, 'player_id')
	)


def downgrade(migrator: SchemaMigrator):
	pass



from pyplanet.apps.config import AppConfig
from pyplanet.apps.contrib.funcmd.view import EmojiToolbarView
from pyplanet.contrib.command import Command

from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.contrib.setting import Setting


class FunCmd(AppConfig):
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.setting_emoji_toolbar = Setting(
			'emoji_toolbar', 'Display Emoji Toolbar', Setting.CAT_DESIGN, type=bool, default=True,
			description='Display the Emoji Toolbar to users.',
			change_target=self.reload_settings
		)

		self.emoji_toolbar = EmojiToolbarView(self)

	async def on_start(self):
		await self.context.setting.register(
			self.setting_emoji_toolbar
		)

		await self.instance.command_manager.register(
			Command(command='afk', target=self.command_afk, admin=False, description='Set yourself as AFK'),
			Command(command='bootme', target=self.command_bootme, admin=False, description='Boot yourself from the server'),
			Command(command='rq', aliases=['ragequit'], target=self.command_rq, admin=False, description='Ragequit from the server'),
			Command(command='gg', target=self.command_gg, admin=False, description='Send Good Game to everyone'),
			Command(command='n1', target=self.command_n1, admin=False, description='Send Nice One to everyone'),
			Command(command='nt', target=self.command_nt, admin=False, description='Send Nice Try/Nice Time to everyone'),
		)

		if self.instance.game.game == 'sm':
			await self.instance.command_manager.register(
				Command(command='ns', target=self.command_ns, admin=False, description='Send Nice Shot to everyone'),
			)

		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)

		if await self.setting_emoji_toolbar.get_value():
			await self.emoji_toolbar.display()

	async def reload_settings(self, *args, **kwargs):
		if await self.setting_emoji_toolbar.get_value():
			await self.emoji_toolbar.display()
		else:
			await self.emoji_toolbar.hide()

	async def player_connect(self, player, *args, **kwargs):
		if await self.setting_emoji_toolbar.get_value():
			await self.emoji_toolbar.display(player_logins=[player.login])

	async def command_afk(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.gbx.multicall(
			self.instance.gbx('ForceSpectator', player.login, 3),
			self.instance.chat('$fff {}$z$s$fff is now away from keyboard.'.format(player.nickname))
		)

	async def command_bootme(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.gbx.multicall(
			self.instance.chat('$fff  {}$z$s$fff chooses to boot back to real life!'.format(player.nickname)),
			self.instance.gbx('Kick', player.login, 'chooses to boot to real life (/bootme)'),
		)

	async def command_rq(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.gbx.multicall(
			self.instance.chat('$f00 {}$z$s$f00 rage quits.'.format(player.nickname)),
			self.instance.gbx('Kick', player.login, 'rage quit (/rq)'),
		)

	async def command_gg(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.chat('$fff {}$z$s$fff Good Game everyone!'.format(player.nickname))

	async def command_n1(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.chat('$fff {}$z$s$fff Nice one!'.format(player.nickname))

	async def command_ns(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.chat('$fff {}$z$s$fff Nice shot!'.format(player.nickname))

	async def command_nt(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		if self.instance.game.game == 'sm':
			await self.instance.chat('$fff {}$z$s$fff Nice try!'.format(player.nickname))
		else:
			await self.instance.chat('$fff {}$z$s$fff Nice time!'.format(player.nickname))





class UIException(Exception):
	"""
	Base exception for UI core component.
	"""
	pass


class ManialinkMemoryLeakException(UIException):
	"""
	Is thrown when a memory leak is detected in a view. Raised when a manialink responds to a view, but the view is 
	vanished for the specified player(s).
	"""
	pass


class UIPropertyDoesNotExist(UIException):
	"""
	Thrown when UI Property with element doesn't exist.
	"""
	pass



"""
Toolbar app component
"""
from pyplanet.apps.core.pyplanet.views.setting import SettingMenuView
from pyplanet.apps.core.pyplanet.views.toolbar import ToolbarView
from pyplanet.contrib.setting import Setting

from pyplanet.apps.core.maniaplanet import callbacks as mp_signals


class ToolbarComponent:
	def __init__(self, app):
		"""
		Initiate toolbar component.

		:param app: App config instance
		:type app: pyplanet.apps.core.pyplanet.app.PyPlanetConfig
		"""
		self.app = app

		self.setting_display_toolbar = Setting(
			'display_player_toolbar', 'Display player toolbar', Setting.CAT_DESIGN, type=bool, default=True,
			description='Display the toolbar to all the players, the toolbar contains some useful buttons',
			change_target=self.reload_settings
		)

		self.widget = ToolbarView(self.app)

	async def on_init(self):
		pass

	async def reload_settings(self, *args, **kwargs):
		if await self.setting_display_toolbar.get_value():
			await self.widget.display()
		else:
			await self.widget.hide()

	async def on_start(self):
		await self.app.context.setting.register(
			self.setting_display_toolbar
		)

		self.app.context.signals.listen(mp_signals.player.player_connect, self.player_connect)

		if await self.setting_display_toolbar.get_value():
			await self.widget.display()

	async def player_connect(self, player, *args, **kwargs):
		if await self.setting_display_toolbar.get_value():
			await self.widget.display(player_logins=[player.login])



import re
import struct

import binascii

STRIP_ALL = dict(letters='wnoitsgz<>', part=r'\$[lh]\[.+\]|\$[lh]|\$[0-9a-f]{3}')
"""
Strip all custom maniaplanet styles + formatting.
"""

STRIP_COLORS = dict(letters='g', part=r'\$[0-9a-f]{3}')
"""
Strip colors from your input (including $g, color reset).
"""

STRIP_SIZES = dict(letters='wnoiz')
"""
Strip all size and adjustments styles ($w $n $o $i $z).
"""

STRIP_SHADOWS = dict(letters='s')
"""
Strip shadow style ($s).
"""

STRIP_CAPITALS = dict(letters='t')
"""
Strip capital style ($t).
"""

STRIP_LINKS = dict(part=r'\$[lh]\[.+\]|\$[lh]')
"""
Strip links ($h and $l).
"""


def style_strip(text, *strip_methods, strip_styling_blocks=True, keep_reset=False, keep_color_reset=False):
	"""
	Strip styles from the Maniaplanet universe.

	Examples:
	
	.. code-block:: python

		print("--- Strip: colours ---")
		print(style_strip("$i$fffMax$06fSmurf$f00.$fffes$$l$09f.$fffm$08f$a5x$n$w$o", STRIP_COLORS))
		print(style_strip("$l[some link]$i$FFFMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx$l", STRIP_COLORS))
		print(style_strip("$l[some link]$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx", STRIP_COLORS))
		print("--- Strip: links ---")
		print(style_strip("$l$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x$l", STRIP_LINKS))
		print(style_strip("$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x", STRIP_LINKS))
		print(style_strip("$l[some link]$i$fffMax$06fSmurf$f00.$fffes$$l$09f.$fffm$08fx$l", STRIP_LINKS))
		print(style_strip("$l[some link]$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx", STRIP_LINKS))
		print("--- Strip: sizes ---")
		print(style_strip("$i$n$fffMax$06fSmurf$f00.$w$o$fffe$$nsl$09f.$w$fffm$08f$a5$ox", STRIP_SIZES))
		print("--- Strip: everything ---")
		print(style_strip("$h$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x$h", STRIP_ALL))
		print(style_strip("$l[some link]$i$fffMax$06fSmur$$f$f00.$fffesl$09f.$fffm$08fx$l"))
		print(style_strip("$l[some link]$i$fffMax$06fSmu$nrf$f00.$fffesl$09f.$fffm$08fx"))
		# Other stuff.:
		print(style_strip("$l[some link]$i$fffMax$06fSmu$nrf$f00.$fffesl$09f.$fffm$08fx", STRIP_CAPITALS, STRIP_SHADOWS))

	.

	:param text: The input string text.
	:param strip_methods: Methods for stripping, use one of the STRIP_* constants or leave undefined to strip everything.
	:param strip_styling_blocks: Strip all styling blocks ($> and $<)
	:param keep_reset: Keep full resets ($z).
	:param keep_color_reset: Keep color resets ($g).
	:type text: str
	:type strip_styling_blocks: bool
	:type keep_reset: bool
	:type keep_color_reset: bool
	:return: Stripped style string.
	:rtype: str
	"""
	if not strip_methods:
		strip_methods = [STRIP_ALL]
	regex = None
	letters = ''
	parts = []
	for payload in strip_methods:
		if isinstance(payload, str):
			regex = payload
			break
		elif isinstance(payload, dict):
			if 'letters' in payload:
				letters += payload['letters']
			if 'part' in payload:
				parts.append(payload['part'])

	if keep_reset:
		letters = letters.replace('z', '')
	if keep_color_reset:
		letters = letters.replace('g', '')
	if strip_styling_blocks:
		letters += '<>'

	if not regex:
		regex = r'(\$[{letters}]{parts})+'.format(
			letters=letters,
			parts='|{}'.format('|'.join(parts)) if len(parts) > 0 else ''
		)

	# Strip and return.
	return re.sub(regex, '', text, flags=re.IGNORECASE)


def percentage_color(percentage):  # pragma: no cover
	red = (255 * percentage) / 100
	green = (255 * (100 - percentage)) / 100
	return int(abs(255-red)), int(abs(255-green)), 0


def rgb_to_hex(rgb):  # pragma: no cover
	return '%02x%02x%02x' % rgb



import stat
import os
import shutil
import sys
import pyplanet

from os import path
from jinja2 import Environment

from pyplanet.conf import settings
from pyplanet.core.management import BaseCommand, CommandError


class TemplateCommand(BaseCommand):
	"""
	Template command is an abstract command to copy template files anywhere.
	"""
	requires_system_checks = False
	requires_migrations_checks = False
	leave_locale_alone = True

	# Define your template files (files that should be templated) here.
	template_files = []

	# Set the template base dir.
	template_base_dir = path.join(pyplanet.__path__[0], 'conf')

	def add_arguments(self, parser):
		parser.add_argument('name', help='Name of folder/app to create.')
		parser.add_argument('directory', nargs='?', help='Optional destination directory. Defaults to new directory.')

	def handle(self, app_or_project, name, target=None, **options):
		self.app_or_project = app_or_project
		self.paths_to_remove = []
		self.verbosity = options['verbosity']

		self.validate_name(name, app_or_project)

		# If it's a relative to the current dir, replace it.
		if name == '.':
			name = ''

		# if some directory is given, make sure it's nicely expanded
		if target is None:
			top_dir = path.join(os.getcwd(), name)
			try:
				os.makedirs(top_dir)
			except FileExistsError:
				listdir = os.listdir(top_dir)
				if len(listdir) == 0 or (len(listdir) == 1 and 'env' in listdir):
					pass
				else:
					raise CommandError("'%s' already exists" % top_dir)
			except OSError as e:
				raise CommandError(e)
		else:
			top_dir = os.path.abspath(path.expanduser(target))
			if not os.path.exists(top_dir):
				raise CommandError("Destination directory '%s' does not "
								   "exist, please create it first." % top_dir)

		if self.verbosity >= 2:
			self.stdout.write("Rendering {} template files\n".format(app_or_project))

		base_name = '{}_name'.format(app_or_project)
		base_subdir = '{}_template'.format(app_or_project)
		base_directory = '{}_directory'.format(app_or_project)
		camel_case_name = 'camel_case_{}_name'.format(app_or_project)
		camel_case_value = ''.join(x for x in name.title() if x != '_')

		context = dict(options, **{
			base_name: name,
			base_directory: top_dir,
			camel_case_name: camel_case_value,
			'pyplanet_version': pyplanet.__version__,
		})

		# Setup a stub settings environment for template rendering
		try:
			if not settings.configured:
				_ = settings.APPS
		except:
			pass

		template_dir = self.handle_template(base_subdir)
		prefix_length = len(template_dir) + 1
		env = Environment(autoescape=False)

		for root, dirs, files in os.walk(template_dir):
			path_rest = root[prefix_length:]
			relative_dir = path_rest.replace(base_name, name)
			if relative_dir:
				target_dir = path.join(top_dir, relative_dir)
				if not path.exists(target_dir):
					os.mkdir(target_dir)

			for dirname in dirs[:]:
				if dirname.startswith('.') or dirname == '__pycache__':
					dirs.remove(dirname)

			for filename in files:
				if filename.endswith(('.pyo', '.pyc', '.py.class')):
					# Ignore some files as they cause various breakages.
					continue
				old_path = path.join(root, filename)
				new_path = path.join(top_dir, relative_dir, filename.replace(base_name, name))

				if path.exists(new_path):
					raise CommandError("%s already exists, overlaying a "
									   "project or app into an existing "
									   "directory won't replace conflicting "
									   "files" % new_path)

				# Only render the Python files, as we don't want to
				# accidentally render PyPlanet templates files
				if path.join(relative_dir, filename) in self.template_files:
					with open(old_path, 'r', encoding='utf-8') as template_file:
						content = template_file.read()
					template = env.from_string(content)
					content = template.render(context)
					with open(new_path, 'w', encoding='utf-8') as new_file:
						new_file.write(content)
				else:
					shutil.copyfile(old_path, new_path)

				if self.verbosity >= 2:
					self.stdout.write("Creating %s\n" % new_path)
				try:
					shutil.copymode(old_path, new_path)
					self.make_writeable(new_path)
				except OSError:
					self.stderr.write(
						"Notice: Couldn't set permission bits on %s. You're "
						"probably using an uncommon filesystem setup. No "
						"problem." % new_path, self.style.NOTICE)

		if self.paths_to_remove:
			if self.verbosity >= 2:
				self.stdout.write("Cleaning up temporary files.\n")
			for path_to_remove in self.paths_to_remove:
				if path.isfile(path_to_remove):
					os.remove(path_to_remove)
				else:
					shutil.rmtree(path_to_remove)

	def validate_name(self, name, app_or_project):
		if name is None:
			raise CommandError("you must provide %s %s name" % (
				"an" if app_or_project == "app" else "a", app_or_project))
		if app_or_project == 'project' and name == '.':
			return True
		# If it's not a valid directory name.
		if not name.isidentifier():
			raise CommandError(
				"%r is not a valid %s name. Please make sure the name is "
				"a valid identifier." % (name, app_or_project)
			)

	def handle_template(self, subdir):
		return path.join(self.template_base_dir, subdir)

	def make_writeable(self, filename):
		"""
		Make sure that the file is writeable.
		Useful if our source is read-only.
		"""
		if sys.platform.startswith('java'):
			# On Jython there is no os.access()
			return
		if not os.access(filename, os.W_OK):
			st = os.stat(filename)
			new_permissions = stat.S_IMODE(st.st_mode) | stat.S_IWUSR
			os.chmod(filename, new_permissions)



"""
The setting contrib contains code for managing and providing settings contexts.
"""

from .setting import Setting
from .manager import GlobalSettingManager

__all__ = [
	'Setting',
	'GlobalSettingManager',
]



"""
Maniaplanet Core Models. This models are used in several apps and should be considered as very stable.
"""
from peewee import *
from pyplanet.core.db import TimedModel
from pyplanet.apps.core.maniaplanet.models import Map, Player


class LocalRecord(TimedModel):
	map = ForeignKeyField(Map, index=True)
	"""
	Map on which the local record is driven.
	"""

	player = ForeignKeyField(Player, index=True)
	"""
	Player who drove the local record.
	"""

	score = IntegerField(
		null=False
	)
	"""
	Time/score of the local record.
	"""

	checkpoints = TextField(
		null=False, default=''
	)
	"""
	List of checkpoints of the local record.
	"""

	class Meta:
		indexes = (
			(('player', 'map'), True),
		)



import os

TEST_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
TEST_FILES_DIR = os.path.join(TEST_DIR, '_files')

async def drop_tables():
	from pyplanet.core import Controller
	instance = Controller.instance
	if instance and instance.db:
		await instance.db.drop_tables()



import importlib
import os

from pyplanet.conf.backends.base import ConfigBackend
from pyplanet.core.exceptions import ImproperlyConfigured


class FileConfigBackend(ConfigBackend):
	name = None
	files = None
	required_files = None

	def __init__(self, **options):
		super().__init__(**options)

		self.directory = None

	def load(self):
		# Make sure we load the defaults first.
		super().load()

		# Make sure we get the directory from the environment variable.
		self.directory = os.environ.get('PYPLANET_SETTINGS_DIRECTORY')

		if not self.directory:
			raise ImproperlyConfigured(
				'Settings directory is not defined! Please define PYPLANET_SETTINGS_DIRECTORY in your '
				'environment or start script (manage.py).'
			)

		# Make directory absolute.
		self.directory = os.path.join(os.getcwd(), self.directory)

		if not os.path.exists(self.directory) or not os.path.isdir(self.directory):
			raise ImproperlyConfigured(
				'Settings directory does not exist or is not a directory! Please define the right PYPLANET_SETTINGS_DIRECTORY '
				'in your environment or start script (manage.py).'
			)

		# Check for the two required files.
		if self.required_files:
			for req_file in self.required_files:
				file_path = os.path.join(self.directory, req_file)
				if not os.path.exists(file_path) or not os.path.isfile(file_path):
					raise ImproperlyConfigured(
						'One of the configuration files doesn\'t exist in the directory: '
						'file: {}'.format(req_file)
					)

		# Add the module itself to the configuration.
		self.settings['SETTINGS_DIRECTORY'] = self.directory

		# The rest should the parent class inherit and override!



"""
:Victory Types:
	1 = time limit reached, 2 = capture, 3 = attacker eliminated, 4 = defenders eliminated.
"""
import asyncio

from pyplanet.core import Controller
from pyplanet.core.events import Callback

async def handle_elite_turn_start(source, signal, **kwargs):
	attacker = await Controller.instance.player_manager.get_player(login=source['attacker'])
	defenders = await asyncio.gather(*[
		Controller.instance.player_manager.get_player(login=p)
		for p in source['defenders']
	])
	return dict(attacker=attacker, defenders=defenders)

async def handle_elite_turn_end(source, signal, **kwargs):
	return dict(victory_type=source['victorytype'])


turn_start = Callback(
	call='Script.Shootmania.Elite.StartTurn',
	namespace='shootmania',
	code='elite_turn_start',
	target=handle_elite_turn_start
)
"""
:Signal: 
	Elite turn start.
:Code:
	``shootmania:elite_turn_start``
:Description:
	Information about the starting turn.
:Original Callback:
	`Script` Shootmania.Elite.StartTurn

:param attacker: Player instance of attacker.
:param defenders: List with player instances of defenders.
:type attacker: pyplanet.apps.core.maniaplanet.models.player.Player
:type defenders: pyplanet.apps.core.maniaplanet.models.player.Player[]
"""


turn_end = Callback(
	call='Script.Shootmania.Elite.EndTurn',
	namespace='shootmania',
	code='elite_turn_end',
	target=handle_elite_turn_end
)
"""
:Signal: 
	Elite turn start.
:Code:
	``shootmania:elite_turn_end``
:Description:
	Information about the ending turn.
:Original Callback:
	`Script` Shootmania.Elite.EndTurn

:param victory_type: Describe how the turn was won. 1 = time limit, 2 = capture, 3 = attacker eliminated, 4 = defenders eliminated
"""



"""
Maniaplanet Core Models. This models are used in several apps and should be considered as very stable.
"""
from peewee import *
from pyplanet.core.db import TimedModel
from pyplanet.utils.functional import empty


class Player(TimedModel):
	LEVEL_PLAYER = 0
	LEVEL_OPERATOR = 1
	LEVEL_ADMIN = 2
	LEVEL_MASTER = 3
	LEVEL_CHOICES = (
		(LEVEL_PLAYER, 'Player'),
		(LEVEL_OPERATOR, 'Operator'),
		(LEVEL_ADMIN, 'Admin'),
		(LEVEL_MASTER, 'MasterAdmin'),
	)

	login = CharField(
		max_length=100,
		null=False,
		index=True,
		unique=True
	)
	"""
	The login of the player will be the unique index we use in the player table.
	"""

	nickname = CharField(max_length=150)
	"""
	The nickname of the player including maniaplanet styles.
	"""

	last_ip = CharField(max_length=46, null=True, default=None)
	"""
	The last used IP-address for the player.
	"""

	last_seen = DateTimeField(null=True, default=None)
	"""
	When is the player last seen on the server.
	"""

	total_playtime = IntegerField(default=0, null=False)
	"""
	The total playtime of the player
	"""

	total_donations = IntegerField(default=0, null=False)
	"""
	The total donations given by the player.
	"""

	level = IntegerField(choices=LEVEL_CHOICES, default=LEVEL_PLAYER)
	"""
	The level of the player. See the LEVEL_CHOICES tuple. Use methods like get_level_string on player object to retrieve
	the name of the level.
	"""

	CACHE = dict()

	def __str__(self):
		return self.login

	def __init__(self, *args, **kwargs):
		self.__flow = PlayerFlow()
		self.__attributes = PlayerAttributes()
		super().__init__(*args, **kwargs)

	@property
	def flow(self):
		"""
		Flow object of the player.

		:return: flow object
		:rtype: pyplanet.apps.core.maniaplanet.models.player.PlayerFlow
		"""
		return self.__flow

	@property
	def attributes(self):
		return self.__attributes

	async def save(self, *args, **kwargs):
		res = await super().save(*args, **kwargs)
		if self.login not in self.CACHE or (self.login in self.CACHE and id(self) != id(self.CACHE[self.login])):
			if self.login in self.CACHE and id(self) != id(self.CACHE[self.login]):
				self.__flow = self.CACHE[self.login].flow
				self.__attributes = self.CACHE[self.login].attributes
			self.CACHE[self.login] = self
		return res

	@classmethod
	async def get_by_login(cls, login, default=empty):
		"""
		Get player by login.

		:param login: Login.
		:param default: Optional default if not exist.
		:return: Player instance
		:rtype: pyplanet.apps.core.maniaplanet.models.player.Player
		"""
		if login in cls.CACHE:
			return cls.CACHE[login]
		try:
			cls.CACHE[login] = player = await cls.get(login=login)
		except DoesNotExist:
			if default is not empty:
				return default
			raise
		return player

	def get_level_string(self):
		for level_nr, level_name in self.LEVEL_CHOICES:
			if self.level == level_nr:
				return level_name
		return None


class PlayerFlow:
	def __init__(self):
		self.in_run = False
		self.player_id = None
		self.team_id = None
		self.is_player = None
		self.is_spectator = None
		self.is_temp_spectator = None
		self.is_pure_spectator = None
		self.spectator_target = None
		self.force_spectator = None
		self.is_referee = None
		self.is_podium_ready = None
		self.is_using_stereoscopy = None
		self.is_managed_by_other_server = None
		self.is_server = None
		self.has_player_slot = None
		self.is_broadcasting = None
		self.has_joined_game = None
		self.zone = None
		self.joined_at = None

		self.royal_server_start_time = 0
		self.royal_total_time = 0
		self.royal_section_time = 0
		self.royal_times = list()
		self.royal_block_ids = list()

		self.other = dict()

	def start_run(self):
		self.in_run = True

	def handle_waypoint_royal(self, race_time, server_time, block_id):
		"""
		Returns true if handled.
		"""
		if block_id in self.royal_block_ids:
			return False
		self.royal_block_ids.append(block_id)

		time = server_time - self.royal_server_start_time

		self.royal_total_time += time
		self.royal_section_time += time
		self.royal_times.append(self.royal_section_time)

		return True

	def reset_royal(self, server_time):
		self.royal_server_start_time = server_time
		self.royal_total_time = 0
		self.royal_section_time = 0
		self.royal_block_ids = list()
		self.royal_times = list()

	def handle_give_up_royal(self, server_time):
		penalty = 1500
		time = server_time - self.royal_server_start_time + penalty
		self.royal_total_time += time
		self.royal_section_time += time

	def handle_start_line_royal(self, server_time):
		self.royal_server_start_time = server_time

	def handle_match_begin_royal(self):
		self.royal_times = list()
		self.royal_block_ids = list()
		self.royal_server_start_time = 0
		self.royal_section_time = 0
		self.royal_total_time = 0

	def reset_run(self):
		if self.in_run:
			self.in_run = False

	def update_state(self, **data):
		self.is_spectator = data.pop('is_spectator')
		self.is_player = not self.is_spectator
		self.spectator_target = data.pop('target')
		self.team_id = data.pop('team_id')

		for key, value in data.items():
			setattr(self, key, value)

	def reset_state(self):
		self.is_player = None
		self.is_spectator = None
		self.is_pure_spectator = None
		self.is_temp_spectator = None
		self.force_spectator = None
		self.is_referee = None
		self.is_podium_ready = None
		self.is_using_stereoscopy = None
		self.is_managed_by_other_server = None
		self.is_server = None
		self.has_player_slot = None
		self.is_broadcasting = None
		self.has_joined_game = None


class PlayerAttributes:
	"""
	Hold custom attributes by keys used in several apps.
	"""
	def __init__(self):
		self.__attributes = dict()

	def set(self, key, value):
		"""
		Set value of a defined key.

		:param key: Key string
		:param value: Value of the attribute
		:type key: str
		"""
		self.__attributes[key] = value

	def get(self, key, default=None):
		"""
		Get value of the key (or the default value).

		:param key: Key string
		:param default: Default value, defaults to None
		:return: The value of the attribute.
		"""
		if key in self.__attributes:
			return self.__attributes[key]
		return default

	def all(self):
		"""
		Get all attributes as dictionary. Please don't change attributes this way as the backend of the attributes may change
		over time!

		:return: Dict of attributes.
		:rtype: dict
		"""
		return self.__attributes



from peewee import DoesNotExist

from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.apps.core.pyplanet.models import Permission
from pyplanet.contrib import CoreContrib


class PermissionManager(CoreContrib):
	"""
	Permission Manager manges the permissions of all apps and players.
	
	.. todo::
	
		Write introduction.
		
	.. warning::
	
		Don't initiate this class yourself.
	"""
	def __init__(self, instance):
		"""
		Initiate, should only be done from the core instance.
		
		:param instance: Instance.
		:type instance: pyplanet.core.instance.Instance
		"""
		self._instance = instance

	async def on_start(self):
		"""
		Handle startup, just before the apps will start. We will make sure we are ready to get requests for permissions.
		"""
		pass

	async def has_permission(self, player, permission):
		"""
		Check if the player has the right permission.
		
		:param player: player instance.
		:param permission: permission name.
		:return: boolean if player is allowed.
		"""
		if isinstance(permission, str):
			perm_namespace, _, perm_name = permission.rpartition(':')
			permission = await self.get_perm(name=perm_name, namespace=perm_namespace)
		if isinstance(player, str):
			player = await self._instance.player_manager.get_player(login=player)
		if not isinstance(permission, Permission):
			raise Exception('Permission should be a string or permission object!')
		if not isinstance(player, Player):
			raise Exception('Player should be a string or player object!')
		return player.level >= permission.min_level

	async def get_perm(self, namespace, name):
		"""
		Get permission by namespace and name.
		
		:param namespace: Namespace of the permission
		:param name: Name of the permission.
		:type name: str
		:type namespace: str
		"""
		return await Permission.get(namespace=namespace, name=name)

	async def register(self, name, description='', app=None, min_level=1, namespace=None):
		"""
		Register a new permission.
		
		:param name: Name of permission
		:param description: Description in english.
		:param app: App instance to retrieve the label.
		:param min_level: Minimum level required.
		:param namespace: Namespace, only for core usage!
		:return: Permission instance.
		"""
		if not namespace and app:
			namespace = app.label
		if not namespace:
			raise Exception('Namespace is required. You should give your app instance with app=app instead!')

		try:
			perm = await self.get_perm(namespace=namespace, name=name)

			# TODO: Implement overrides on min_level here.
			if perm.min_level != min_level:
				perm.min_level = min_level
				await perm.save()

		except DoesNotExist:
			perm = Permission(namespace=namespace, name=name, description=description, min_level=min_level)
			await perm.save()
		return perm



"""
:Weapons:
	[1-Laser, 2-Rocket, 3-Nucleus, 5-Arrow]
"""
import asyncio

from pyplanet.core import Controller
from pyplanet.core.events import Callback, handle_generic


async def _get_player(login):
	try:
		if not login or not len(login):
			return None
		return await Controller.instance.player_manager.get_player(login=login)
	except:
		return None


async def handle_on_shoot(source, signal, **kwargs):
	shooter = await Controller.instance.player_manager.get_player(login=source['shooter'])
	return dict(shooter=shooter, time=source['time'], weapon=source['weapon'])

async def handle_on_hit(source, signal, **kwargs):
	victim, shooter = await asyncio.gather(*[
		_get_player(source['victim']),
		_get_player(source['shooter']),
	])
	return dict(
		shooter=shooter, time=source['time'], weapon=source['weapon'], victim=victim, damage=source['damage'],
		points=source['points'], distance=source['distance'],
		shooter_position=source['shooterposition'], victim_position=source['victimposition'],
	)

async def handle_on_near_miss(source, signal, **kwargs):
	victim, shooter = await asyncio.gather(*[
		_get_player(source['victim']),
		_get_player(source['shooter']),
	])
	return dict(
		shooter=shooter, time=source['time'], weapon=source['weapon'], victim=victim, distance=source['distance'],
		shooter_position=source['shooterposition'], victim_position=source['victimposition'],
	)

async def handle_armor_empty(source, signal, **kwargs):
	victim, shooter = await asyncio.gather(
		_get_player(source['victim']),
		_get_player(source['shooter']),
	)

	return dict(
		shooter=shooter, time=source['time'], weapon=source['weapon'], victim=victim, distance=source['distance'],
		shooter_position=source['shooterposition'], victim_position=source['victimposition'],
	)

async def handle_on_capture(source, signal, **kwargs):
	players = await asyncio.gather(*[
		Controller.instance.player_manager.get_player(login=p)
		for p in source['players']
	])
	return dict(time=source['time'], players=players, landmark=source['landmark'])

async def handle_on_shot_deny(source, signal, **kwargs):
	victim, shooter = await asyncio.gather(*[
		_get_player(source['victim']),
		_get_player(source['shooter']),
	])
	return dict(
		time=source['time'], shooter=shooter, victim=victim,
		shooter_weapon=source['shooterweapon'], victim_weapon=source['victimweapon']
	)

async def handle_on_fall_damage(source, signal, **kwargs):
	victim = await Controller.instance.player_manager.get_player(login=source['victim'])
	return dict(time=source['time'], victim=victim)

async def handle_player_added(source, signal, **kwargs):
	player = await Controller.instance.player_manager.get_player(login=source['login'])
	return dict(
		time=source['time'], player=player, team=source['team'], language=source['language'],
		ladder_rank=source['ladderrank'], ladder_points=source['ladderpoints']
	)

async def handle_custom_event(source, signal, **kwargs):
	victim, shooter = await asyncio.gather(*[
		_get_player(source['victim']),
		_get_player(source['shooter']),
	])
	source['shooter'] = shooter
	source['victim'] = victim
	return source

async def handle_action_event(source, signal, **kwargs):
	player = await _get_player(login=source['login'])
	return dict(time=source['time'], player=player, action_input=source['actioninput'])

async def handle_player_touches_object(source, signal, **kwargs):
	player = await _get_player(login=source['login'])
	return dict(
		time=source['time'], player=player, object_id=source['objectid'], model_id=source['modelid'], model_name=source['modelname']
	)

async def handle_player_triggers_sector(source, signal, **kwargs):
	player = await _get_player(login=source['login'])
	return dict(time=source['time'], player=player, sector_id=source['sectorid'])

async def handle_player_request_action_change(source, signal, **kwargs):
	player = await _get_player(login=source['login'])
	return dict(time=source['time'], player=player, action_change=source['actionchange'])

async def handle_scores(source, signal, **kwargs):
	async def get_player_scores(data):
		player = await _get_player(login=data['login'])
		return dict(
			player=player, rank=data['rank'], round_points=data['roundpoints'], map_points=data['mappoints'],
			match_points=data['matchpoints'],
		)
	async def get_team_scores(data):
		return dict(
			name=data['name'], id=data['id'], round_points=data['roundpoints'],
			map_points=data['mappoints'], match_points=data['matchpoints'],
		)

	player_scores = await asyncio.gather(*[
		get_player_scores(d) for d in source['players']
	])
	team_scores = await asyncio.gather(*[
		get_team_scores(d) for d in source['teams']
	])
	return dict(
		players=player_scores,
		teams=team_scores,
		winner_team=source['winnerteam'],
		use_teams=source['useteams'],
		winner_player=source['winnerplayer'],
		section=source['section']
	)

#######################################################################################
# Generic
#######################################################################################

on_shoot = Callback(
	call='Script.Shootmania.Event.OnShoot',
	namespace='shootmania',
	code='on_shoot',
	target=handle_on_shoot,
)
"""
:Signal: 
	Player shoot.
:Code:
	``shootmania:on_shoot``
:Description:
	Callback sent when a player shoots.
:Original Callback:
	`Script` Shootmania.Event.OnShoot

:param shooter: Shooter, Player instance
:param time: Time of server when callback is sent.
:param weapon: Weapon number.
:type shooter: pyplanet.apps.core.maniaplanet.models.player.Player
"""

on_hit = Callback(
	call='Script.Shootmania.Event.OnHit',
	namespace='shootmania',
	code='on_hit',
	target=handle_on_hit,
)
"""
:Signal: 
	Player hit.
:Code:
	``shootmania:on_hit``
:Description:
	Callback sent when a player is hit.
:Original Callback:
	`Script` Shootmania.Event.OnHit

:param shooter: shooter, Player instance
:param time: Time of server when callback is sent.
:param weapon: Weapon number.
:param victim: victim, Player instance
:param damage: Damage done.
:param points: Points scored by hit.
:param distance: Distance between victim and shooter.
:param shooter_position: Position of shooter.
:param victim_position: Position of victim.
:type shooter: pyplanet.apps.core.maniaplanet.models.player.Player
:type victim: pyplanet.apps.core.maniaplanet.models.player.Player
"""

on_near_miss = Callback(
	call='Script.Shootmania.Event.OnNearMiss',
	namespace='shootmania',
	code='on_near_miss',
	target=handle_on_near_miss,
)
"""
:Signal: 
	Near Miss.
:Code:
	``shootmania:on_near_miss``
:Description:
	Callback sent when a player dodges a projectile.
:Original Callback:
	`Script` Shootmania.Event.OnNearMiss

:param shooter: shooter, Player instance
:param time: Time of server when callback is sent.
:param weapon: Weapon number.
:param victim: victim, Player instance
:param distance: Distance between victim and shooter.
:param shooter_position: Position of shooter.
:param victim_position: Position of victim.
:type shooter: pyplanet.apps.core.maniaplanet.models.player.Player
:type victim: pyplanet.apps.core.maniaplanet.models.player.Player
"""

on_armor_empty = Callback(
	call='Script.Shootmania.Event.OnArmorEmpty',
	namespace='shootmania',
	code='on_armor_empty',
	target=handle_armor_empty,
)
"""
:Signal: 
	Armor empty, player eliminated.
:Code:
	``shootmania:on_armor_empty``
:Description:
	Callback sent when a player is eliminated.
:Original Callback:
	`Script` Shootmania.Event.OnArmorEmpty

:param shooter: shooter, Player instance
:param time: Time of server when callback is sent.
:param weapon: Weapon number.
:param victim: victim, Player instance
:param distance: Distance between victim and shooter.
:param shooter_position: Position of shooter.
:param victim_position: Position of victim.
:type shooter: pyplanet.apps.core.maniaplanet.models.player.Player
:type victim: pyplanet.apps.core.maniaplanet.models.player.Player
"""


on_capture = Callback(
	call='Script.Shootmania.Event.OnCapture',
	namespace='shootmania',
	code='on_capture',
	target=handle_on_capture,
)
"""
:Signal: 
	Landmark has been captured
:Code:
	``shootmania:on_capture``
:Description:
	Callback sent when a landmark is captured.
:Original Callback:
	`Script` Shootmania.Event.OnCapture

time=source['time'], players=players, landmark=source['landmark']

:param time: Time of server when callback is sent.
:param players: Player list (instances).
:param landmark: Landmark information, raw!
:type players: pyplanet.apps.core.maniaplanet.models.player.Player[]
"""


on_shot_deny = Callback(
	call='Script.Shootmania.Event.OnShotDeny',
	namespace='shootmania',
	code='on_shot_deny',
	target=handle_on_shot_deny,
)
"""
:Signal: 
	Player denies a projectile.
:Code:
	``shootmania:on_shot_deny``
:Description:
	Callback sent when a player denies a projectile.
:Original Callback:
	`Script` Shootmania.Event.OnShotDeny

:param time: Time of server when callback is sent.
:param shooter: shooter, Player instance
:param victim: victim, Player instance
:param shooter_weapon: Weapon number of shooter.
:param victim_weapon: Weapon number of victim that denied the shot.
:param distance: Distance between victim and shooter.
:param shooter_position: Position of shooter.
:param victim_position: Position of victim.
:type shooter: pyplanet.apps.core.maniaplanet.models.player.Player
:type victim: pyplanet.apps.core.maniaplanet.models.player.Player
"""

on_fall_damage = Callback(
	call='Script.Shootmania.Event.OnFallDamage',
	namespace='shootmania',
	code='on_fall_damage',
	target=handle_on_fall_damage,
)
"""
:Signal: 
	Fall Damage
:Code:
	``shootmania:on_fall_damage``
:Description:
	Callback sent when a player suffers fall damage.
:Original Callback:
	`Script` Shootmania.Event.OnFallDamage

:param time: Time of server when callback is sent.
:param victim: victim, Player instance
:type victim: pyplanet.apps.core.maniaplanet.models.player.Player
"""


on_command = Callback(
	call='Script.Shootmania.Event.OnCommand',
	namespace='shootmania',
	code='on_command',
	target=handle_generic,
)
"""
:Signal: 
	On Command
:Code:
	``shootmania:on_command``
:Description:
	Callback sent when a command is executed on the server.
:Original Callback:
	`Script` Shootmania.Event.OnCommand

:param time: Time of server when callback is sent.
:param name: Name of the command
:param value: Value in dictionary of the command.
:type value: dict
"""


on_default = Callback(
	call='Script.Shootmania.Event.Default',
	namespace='shootmania',
	code='on_default',
	target=handle_generic,
)
"""
:Signal: 
	On Default Event
:Code:
	``shootmania:on_default``
:Description:
	Callback sent when a old event or default event has been fired.
:Original Callback:
	`Script` Shootmania.Event.Default

:param time: Time of server when callback is sent.
:param type: Name of the command
"""


player_added = Callback(
	call='Script.Shootmania.Event.OnPlayerAdded',
	namespace='shootmania',
	code='player_added',
	target=handle_player_added,
)
"""
:Signal: 
	On player added.
:Code:
	``shootmania:player_added``
:Description:
	Callback sent when a player joins the server.
:Original Callback:
	`Script` Shootmania.Event.OnPlayerAdded

:param time: Time of server when callback is sent.
:param player: Player instance
:param team: Team nr.
:param language: Language code, like 'en'.
:param ladder_rank: Current ladder rank.
:param ladder_points: Current ladder points.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


player_removed = Callback(
	call='Script.Shootmania.Event.OnPlayerRemoved',
	namespace='shootmania',
	code='player_removed',
	target=handle_generic,
)
"""
:Signal: 
	On player removed.
:Code:
	``shootmania:player_removed``
:Description:
	Callback sent when a player leaves the server.
:Original Callback:
	`Script` Shootmania.Event.OnPlayerRemoved

:param time: Time of server when callback is sent.
:param login: Player login string
:param player: Player instance.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


player_request_respawn = Callback(
	call='Script.Shootmania.Event.OnPlayerRequestRespawn',
	namespace='shootmania',
	code='player_request_respawn',
	target=handle_generic,
)
"""
:Signal: 
	On player request respawn.
:Code:
	``shootmania:player_request_respawn``
:Description:
	Callback sent when a player presses the respawn button.
:Original Callback:
	`Script` Shootmania.Event.OnPlayerRequestRespawn

:param time: Time of server when callback is sent.
:param login: Player login string
:param player: Player instance.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


action_custom_event = Callback(
	call='Script.Shootmania.Event.OnActionCustomEvent',
	namespace='shootmania',
	code='action_custom_event',
	target=handle_custom_event,
)
"""
:Signal: 
	Handle Action Custom Event.
:Code:
	``shootmania:action_custom_event``
:Description:
	Callback sent when an action triggers a custom event.
:Original Callback:
	`Script` Shootmania.Event.OnActionCustomEvent

:param time: Time of server when callback is sent.
:param shooter: Shooter player instance if any
:param victim: Victim player instance if any
:param actionid: Action Identifier.
:param *: Any other params, like ``param1``, ``param2``, etc...
:type shooter: pyplanet.apps.core.maniaplanet.models.player.Player
:type victim: pyplanet.apps.core.maniaplanet.models.player.Player
"""


action_event = Callback(
	call='Script.Shootmania.Event.OnActionEvent',
	namespace='shootmania',
	code='action_event',
	target=handle_action_event,
)
"""
:Signal: 
	Handle Action Event.
:Code:
	``shootmania:action_event``
:Description:
	Callback sent when an action triggers an event.
:Original Callback:
	`Script` Shootmania.Event.OnActionEvent

:param time: Time of server when callback is sent.
:param login: Player login
:param player: Player instance.
:param action_input: Action input.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


player_touches_object = Callback(
	call='Script.Shootmania.Event.OnPlayerTouchesObject',
	namespace='shootmania',
	code='player_touches_object',
	target=handle_player_touches_object,
)
"""
:Signal: 
	Player Touches Object.
:Code:
	``shootmania:player_touches_object``
:Description:
	Callback sent when a player touches an object.
:Original Callback:
	`Script` Shootmania.Event.OnPlayerTouchesObject

:param time: Time of server when callback is sent.
:param player: Player instance.
:param object_id: Object Identifier.
:param model_id: Model identifier.
:param model_name: Model name.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


player_triggers_sector = Callback(
	call='Script.Shootmania.Event.OnPlayerTriggersSector',
	namespace='shootmania',
	code='player_triggers_sector',
	target=handle_player_triggers_sector,
)
"""
:Signal: 
	Player Triggers Sector.
:Code:
	``shootmania:player_triggers_sector``
:Description:
	Callback sent when a player triggers a sector.
:Original Callback:
	`Script` Shootmania.Event.OnPlayerTriggersSector

:param time: Time of server when callback is sent.
:param player: Player instance.
:param sector_id: Sector Identifier.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


player_throws_object = Callback(
	call='Script.Shootmania.Event.OnPlayerThrowsObject',
	namespace='shootmania',
	code='player_throws_object',
	target=handle_player_touches_object,  # Can be reused!
)
"""
:Signal: 
	Player Throws an object.
:Code:
	``shootmania:player_touch_object``
:Description:
	Callback sent when a player throws an object.
:Original Callback:
	`Script` Shootmania.Event.OnPlayerThrowsObject

:param time: Time of server when callback is sent.
:param player: Player instance.
:param object_id: Object Identifier.
:param model_id: Model identifier.
:param model_name: Model name.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


player_request_action_change = Callback(
	call='Script.Shootmania.Event.OnPlayerRequestActionChange',
	namespace='shootmania',
	code='player_request_action_change',
	target=handle_player_request_action_change,
)
"""
:Signal: 
	Player requests action change.
:Code:
	``shootmania:player_request_action_change``
:Description:
	Callback sent when a player requests to use another action.
:Original Callback:
	`Script` Shootmania.Event.OnPlayerRequestActionChange

:param time: Time of server when callback is sent.
:param player: Player instance.
:param action_change: Can be -1 (request previous action) or 1 (request next action)
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


scores = Callback(
	call='Script.Shootmania.Scores',
	namespace='shootmania',
	code='scores',
	target=handle_scores
)
"""
:Signal: 
	Score callback, called after the map. (Around the podium time).
:Code:
	``shootmania:scores``
:Description:
	Teams and players scores.
:Original Callback:
	`Script` Shootmania.Scores

:param players: Player score payload. Including player instance etc.
:param teams: Team score payload.
:param winner_team: The winning team.
:param use_teams: Use teams.
:param winner_player: The winning player.
:param section: Section, current progress of match. Important to check before you save results!!
:type players: list
:type teams: list
"""



import asyncio

from pyplanet.core import Controller
from pyplanet.core.events import Callback, handle_generic


async def handle_joust_selected_players(source, signal, **kwargs):
	players = await asyncio.gather(*[
		Controller.instance.player_manager.get_player(login=p)
		for p in source['players']
	])
	return dict(players=players)

async def handle_joust_results(source, signal, **kwargs):
	async def get_player_result(data):
		player = await Controller.instance.player_manager.get_player(login=data['login'])
		return dict(player=player, score=data['score'])
	players = await asyncio.gather(*[
		get_player_result(d) for d in source['players']
	])
	return dict(players=players)


player_reload = Callback(
	call='Script.Shootmania.Joust.OnReload',
	namespace='shootmania',
	code='joust_player_reload',
	target=handle_generic
)
"""
:Signal: 
	Player reloads its weapon and capture pole.
:Code:
	``shootmania:joust_player_reload``
:Description:
	Callback sent when a player capture a pole to reload its weapons.
:Original Callback:
	`Script` Shootmania.Joust.OnReload

:param login: Player login.
:param player: Player instance.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


selected_players = Callback(
	call='Script.Shootmania.Joust.SelectedPlayers',
	namespace='shootmania',
	code='joust_selected_players',
	target=handle_joust_selected_players
)
"""
:Signal: 
	Round starts with selected players.
:Code:
	``shootmania:joust_selected_players``
:Description:
	Callback sent at the beginning of the round with the logins of the players selected to play the round.
:Original Callback:
	`Script` Shootmania.Joust.SelectedPlayers

:param players: Player list (instances).
:type players: pyplanet.apps.core.maniaplanet.models.player.Player[]
"""


results = Callback(
	call='Script.Shootmania.Joust.RoundResult',
	namespace='shootmania',
	code='joust_results',
	target=handle_joust_results
)
"""
:Signal: 
	End of round with results of Joust round.
:Code:
	``shootmania:joust_results``
:Description:
	Callback sent at the end of the round with the scores of the two players.
:Original Callback:
	`Script` Shootmania.Joust.RoundResult

:param players: Player score list, contains player + score.
:type players: list
"""



from getpass import getpass

from pyplanet.contrib.converter import get_converter
from pyplanet.core import Controller
from pyplanet.core.management import BaseCommand


class Command(BaseCommand):
	help = 'Convert the database from XAseco2, eXpansion, ManiaControl and other formats to PyPlanet.'

	requires_migrations_checks = True
	requires_system_checks = True

	arg_pool_required = True
	arg_settings_required = True

	def add_arguments(self, parser):
		parser.add_argument(
			'--source-format', required=True, help='The source format/controller type.',
			choices=['xaseco2', 'uaseco', 'maniacontrol', 'expansion'],
		)
		parser.add_argument(
			'--source-db-type', default='mysql', help='The source database type.',
			choices=['mysql'],
		)
		parser.add_argument('--source-db-username', help='Source database username.', required=True)
		parser.add_argument('--source-db-host', help='Source database hostname.', default='localhost')
		parser.add_argument('--source-db-name', help='Source database name (schema/database)', required=True)
		parser.add_argument('--source-db-port', help='Source database port, leave empty for the default one', default=None)
		parser.add_argument('--source-db-password', help='Source database password. Leave empty for asking', default=None)
		parser.add_argument(
			'--source-db-prefix',
			help='Source database table prefix. Leave empty for using no prefix or the default one by the source type.',
			default=None,
		)

	def handle(self, *args, **options):
		if options['source_db_password'] is None:
			options['source_db_password'] = getpass('Database Password: ')

		instance = Controller.prepare(options['pool']).instance
		converter = get_converter(
			options['source_format'], instance=instance, db_name=options['source_db_name'],
			db_type=options['source_db_type'], db_user=options['source_db_username'],
			db_port=options['source_db_port'], db_password=options['source_db_password'],
			db_host=options['source_db_host'], prefix=options['source_db_prefix']
		)

		instance.loop.run_until_complete(self.convert(instance, converter))

	async def convert(self, instance, converter):
		await converter.connect()
		await converter.start()



import asyncio
import logging
import re

from pyplanet.core.gbx.query import Query, ScriptQuery
from pyplanet.utils.functional import empty
from .remote import GbxRemote

logger = logging.getLogger(__name__)


class GbxClient(GbxRemote):
	"""
	The GbxClient implements and extends the GbxRemote (very base) with the initializers, special encoders/decoders,
	some fixes/hacks needed for the gbx protocol and some tweaks. This is all to prevent changes from the core
	most important logic `GbxRemote`.
	"""

	AUTO_RESPONSE_ID = object()
	SUPPORTED_SCRIPT_API_VERSIONS = [
		'2.3.0', '2.4.0', '2.5.0', '2.6.0', '3.0.0', '3.1.0', '3.2.0', '3.3.0'
	]
	MINIMUM_DEDICATED_VERSION = ['2018', '02', '09',
								 '16', '00']

	def __init__(self, *args, script_api_version=empty, **kwargs):
		super().__init__(*args, **kwargs)

		self.script_api_version = self.SUPPORTED_SCRIPT_API_VERSIONS[len(self.SUPPORTED_SCRIPT_API_VERSIONS)-1]
		if script_api_version != empty and isinstance(script_api_version, str):
			self.script_api_version = script_api_version

		self.game = self.instance.game
		self.refresh_task = None

	def __call__(self, *args, **kwargs):
		if len(args) <= 0:
			return
		return self.prepare(args[0], *args[1:], **kwargs)

	def prepare(self, method, *args, **kwargs):
		"""
		Prepare an query.

		:param method: Method name
		:param args: Arguments...
		:return: Prepared query.
		:rtype: pyplanet.core.gbx.query.Query
		"""
		if method not in self.gbx_methods:
			return ScriptQuery(self, method, *args, **kwargs)
		return Query(self, method, *args, **kwargs)

	async def script(self, method, *args, encode_json=True, response_id=True):
		"""
		Execute scripted call.

		:param method: Scripted method name
		:param args: Arguments
		:param encode_json: Are the arguments dictionary, should it be encoded? (Then only provide the first arg).
		:param response_id: Does the call work on response_id's?
							True by default, set to false to not expect response with response_id
		:return: Future.
		"""
		query = ScriptQuery(self, method, *args, encode_json=encode_json, response_id=response_id)
		return await query.execute()

	async def multicall(self, *queries):
		"""
		Run the queries given async. Will use one or more multicall(s), depends on content.

		:param queries: Queries to execute in multicall.
		:return: Results in tuple.
		:rtype: tuple<any>
		"""
		if len(queries) == 0:
			return tuple()

		# If we got a list instead. unpack it.
		# if len(queries) == 1 and isinstance(queries[0], collections.Iterable):
		# 	queries = queries[0]

		# We will try to put the maximum possible calls into one multicall, for this we need to calculate the lengths
		# so we can stay under the maximum allowed package size.
		current_length = 0

		# Current stack holds the current multicall,
		# will be rotated once the multicall reaches the maximum request size.
		current_stack = list()

		# We will stack calls into multicalls. Structure of this list:
		# multicalls = list(   list()   ) = The list contains lists with calls.
		multicalls = list()

		# Create the multicall(s)
		for query in queries:
			if not isinstance(query, Query):
				continue

			query.prepare()
			if current_length + (query.length + 8) < self.MAX_REQUEST_SIZE:
				current_stack.append(query)
				current_length += (query.length + 8)
			else:
				multicalls.append(current_stack)
				current_length = 0
				current_stack = list()

		# Append the last stack.
		if len(current_stack) > 0:
			multicalls.append(current_stack)

		# Create multicall queries.
		calls = list()
		for mc in multicalls:
			calls.append(
				self.execute('system.multicall', [{'methodName': c.method, 'params': c.args} for c in mc])
			)

		multi_results = await asyncio.gather(*calls)
		results = list()
		for res in multi_results:
			if isinstance(res, list) and len(res) > 0 and isinstance(res[0], list):
				# When we have a list inside our list, we will unwrap that to our root results. This is mostly the case,
				# except when we got error messages!
				for row in res:
					results += row
			else:
				results += res

		return results

	async def connect(self):
		await super().connect()
		await self.initialize()
		await self.instance.storage.initialize()

	async def initialize(self):
		"""
		The initialize method will gather information about the server that is only fetched once and saved and used
		in some classes/parts of the application. To make this information available, we use the pyplanet.core.state
		class instance which is a singleton class instance holding information about the server that is public and
		should be stable.
		"""
		# Verify the dedicated version.
		if self.dedicated_build:
			# Try to parse and verify version.
			matches_minimum_version = False
			try:
				matches = re.findall('(\d+)', self.dedicated_build)
				for part_nr, part_version in enumerate(matches):
					if part_nr <= len(self.MINIMUM_DEDICATED_VERSION):
						if int(part_version) > int(self.MINIMUM_DEDICATED_VERSION[part_nr]):
							matches_minimum_version = True
							break
						elif part_version == self.MINIMUM_DEDICATED_VERSION[part_nr]:
							if part_nr == len(self.MINIMUM_DEDICATED_VERSION) - 1:
								matches_minimum_version = True
								break
						else:
							matches_minimum_version = False
							break
			except:
				pass

			if not matches_minimum_version:
				logger.error(
					'Current dedicated version seems to be not supported by the current PyPlanet version! Minimum: {}_{}'.format(
						'-'.join(self.MINIMUM_DEDICATED_VERSION[:3]),
						'_'.join(self.MINIMUM_DEDICATED_VERSION[3:])
					)
				)
				await asyncio.sleep(5)

		# Try to get the script api_versions.
		try:
			api_versions = await self('XmlRpc.GetAllApiVersions', timeout=5)
			if 'versions' in api_versions and self.script_api_version in api_versions['versions']:
				await self('XmlRpc.SetApiVersion', self.script_api_version, response_id=False)
			self.script_api_version = await self('XmlRpc.GetApiVersion')
		except Exception as e:
			logger.error('Can\'t set the script API Version! {}'.format(str(e)))

		await self.refresh_info()

		# Clear the previous created Manialinks.
		await self('SendHideManialinkPage')

		# Schedule refresh every minute!
		self.refresh_task = asyncio.ensure_future(self.__refresh_info_call())

	async def __refresh_info_call(self):
		while True:
			await asyncio.sleep(60)
			await self.refresh_info()

	async def refresh_info(self):
		# Version Information
		res = await self.multicall(
			self('GetVersion'),
			self('GetSystemInfo'),
			self('GameDataDirectory'),
			self('GetMapsDirectory'),
			self('GetSkinsDirectory'),
			self('GetCurrentMapInfo'),
			self('GetServerPassword'),
			self('GetServerPasswordForSpectator'),
			self('GetMaxPlayers'),
			self('GetMaxSpectators'),
			self('GetHideServer'),
			self('GetLadderServerLimits'),
		)
		version_info = res[0]
		self.game.dedicated_version = version_info['Version']
		self.game.dedicated_build = version_info['Build']
		self.game.dedicated_api_version = version_info['ApiVersion']
		self.game.dedicated_title = version_info['TitleId']

		# System Information
		system_info = res[1]
		self.game.server_is_dedicated = system_info['IsDedicated']
		self.game.server_is_server = system_info['IsServer']
		self.game.server_ip = system_info['PublishedIp']
		self.game.server_p2p_port = system_info['P2PPort']
		self.game.server_port = system_info['Port']
		self.game.server_player_login = system_info['ServerLogin']
		self.game.server_player_id = system_info['ServerPlayerId']
		self.game.server_download_rate = system_info['ConnectionDownloadRate']
		self.game.server_upload_rate = system_info['ConnectionUploadRate']

		self.game.server_data_dir = res[2]
		self.game.server_map_dir = res[3]
		self.game.server_skin_dir = res[4]

		self.game.game = self.game.game_from_environment(
			res[5]['Environnement'],
			game_name=res[0]['Name'],
			title_id=res[0]['TitleId'],
		)

		self.game.server_password = res[6]
		self.game.server_spec_password = res[7]
		self.game.server_max_players = res[8]['CurrentValue']
		self.game.server_next_max_players = res[8]['NextValue']
		self.game.server_max_specs = res[9]['CurrentValue']
		self.game.server_next_max_specs = res[9]['NextValue']
		self.game.server_is_private = res[10]

		self.game.ladder_min = res[11]['LadderServerLimitMin']
		self.game.ladder_max = res[11]['LadderServerLimitMax']

		# Detailed server player infos.
		server_player_info = await self('GetDetailedPlayerInfo', self.game.server_player_login)

		self.game.server_language = server_player_info['Language']
		self.game.server_name = server_player_info['NickName']
		self.game.server_path = server_player_info['Path']


async def multicall(*calls):
	"""
	Run the queries given async. Will use one or more multicall(s), depends on content.

	:param queries: Queries to execute in multicall.
	:return: Results in tuple.
	:rtype: tuple<any>
	"""
	from pyplanet.core import Controller
	return await Controller.instance.gbx.multicall(*calls)




"""
This folder contains the configuration in different files. This init file imports all files in the current directory
to make sure we combine configurations to one settings module.
"""

from .base import *
from .apps import *

try:
	from .local import *
except:
	pass



import textwrap

from pyplanet.contrib import CoreContrib
from pyplanet.contrib.command.command import Command


class CommandManager(CoreContrib):
	"""
	The Command Manager contributed extension is a manager that controls all chat-commands in the game.
	Your app needs to use this manager to register any custom commands you want to provide.

	You should access this class within your app like this:

	.. code-block:: python

		self.instance.command_manager

	You can register your commands like this:

	.. code-block:: python

		await self.instance.command_manager.register(
			Command(command='reboot', target=self.reboot_pool, perms='admin:reboot', admin=True),
		)

	More information of the command and the options of it, see the :class:`pyplanet.contrib.command.Command` class.

	.. warning::

		Don't initiate this class yourself. Access this class from the ``self.instance.command_manager`` instance.

	"""

	def __init__(self, instance):
		"""
		Init manager.

		:param instance: Controller Instance
		:type instance: pyplanet.core.instance.Instance
		"""
		self._instance = instance

		self._commands = list()

	async def on_start(self, **kwargs):
		# Register events.
		self._instance.signals.listen('maniaplanet:player_chat', self._on_chat)

	async def register(self, *commands):
		"""
		Register your command.

		:param commands: Command instance.
		:type commands: pyplanet.contrib.command.command.Command
		"""
		self._commands.extend(commands)

	async def execute(self, player, command, *args):
		"""
		Execute a command for the given player with the given args.

		:param player: Player instance.
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		:param command: Command instance.
		:type command: pyplanet.contrib.command.command.Command
		:param args: Args for the command, will be concat into a string with spaces.
		:return:
		"""
		if isinstance(command, Command):
			command_text = '//' if command.admin else '/'
			if command.namespace:
				command_text += command.namespace + ' '
			command_text += command.command
		else:
			command_text = command

		return await self._on_chat(player, ' '.join([command_text] + list(args)), True)

	async def _on_chat(self, player, text, cmd, **kwargs):
		# Only take action if the chat entry is a command.
		if not cmd:
			return

		# Parse command.
		argv = text.split(' ')
		if not argv:
			return

		# Replace the / in the first part.
		argv[0] = argv[0][1:]

		# Check if we need to ignore the command.
		if len(argv) > 0 and argv[0] in ['serverlogin']:
			return

		# Try to match the command prefix by one of the registered commands.
		command = None
		for cmd in self._commands:
			if cmd.match(argv):
				command = cmd
				break

		# Let the command handle the logic it needs.
		if command:
			return await command.handle(self._instance, player, argv)
		# Send command not found message.
		await self._instance.chat(
			'$z$sCommand unknown. For all commands type /help or //help. '
			'Powered by $l[http://pypla.net]$FD4Py$369Planet',
			player.login
		),

	async def get_command_by_command_text(self, command):
		"""
		Get command by command text. (Used in the /help command)

		:param command: Command entry, array of strings (split by spaces).
		:return: Command object.
		"""
		# Find the right command.
		for cmd in self._commands:
			if cmd.match(command):
				return cmd
		return None

	async def help_entries(self, player, admin_only):  # pragma: no cover
		"""
		Get all help entries for the player.

		:param player: Player instance.
		:param admin_only: Only the admin commands or non-admin. True for admin only, False for player only.
						Will filter on permissions of the player as well!
		:return: List of commands objects.
		"""
		# All commands.
		commands = [c for c in self._commands if c.admin is admin_only]
		if admin_only:
			commands = [c for c in commands if await c.has_permission(self._instance, player)]

		return commands



import datetime
from peewee import *

from pyplanet.apps.core.maniaplanet.models import Map, Player
from pyplanet.core.db import Model


class Score(Model):
	map = ForeignKeyField(Map, index=True)
	"""
	The map the player did play.
	"""

	player = ForeignKeyField(Player, index=True)
	"""
	Player that has driven this time.
	"""

	score = IntegerField()
	"""
	Score (points or time).
	"""

	checkpoints = TextField()
	"""
	The checkpoint times, separated by commas.
	"""

	created_at = DateTimeField(
		default=datetime.datetime.now,
	)
	"""
	When is the time driven?
	"""

	class Meta:
		db_table = 'stats_scores'



"""
Toolbar View.
"""

from pyplanet.views import TemplateView


class ToolbarView(TemplateView):
	"""
	Toolbar View Class.
	"""
	template_name = 'admin/toolbar/toolbar.xml'

	def __init__(self, app):
		"""
		Initiate the toolbar view.

		:param app: Admin app instance.
		:type app: pyplanet.apps.contrib.admin.Admin
		"""
		super().__init__(app.context.ui)

		self.app = app
		self.id = 'admin_toolbar'

		self.subscribe('bar_button_prev', self.action_prev)
		self.subscribe('bar_button_endround', self.action_endround)
		self.subscribe('bar_button_restart', self.action_restart)
		self.subscribe('bar_button_replay', self.action_replay)
		self.subscribe('bar_button_skip', self.action_skip)

		self.subscribe('bar_button_settings', self.action_settings)
		self.subscribe('bar_button_modesettings', self.action_modesettings)
		self.subscribe('bar_button_players', self.action_player_list)

	async def get_context_data(self):
		data = await super().get_context_data()
		data['game'] = self.app.instance.game.game
		return data

	async def action_prev(self, player, *args, **kwargs):
		return await self.app.instance.command_manager.execute(player, '//prev')

	async def action_endround(self, player, *args, **kwargs):
		if self.app.instance.game.game == 'sm':
			return await self.app.instance.chat('$ff0Error: Can\'t end round in Shootmania!', player)
		return await self.app.instance.command_manager.execute(player, '//endround')

	async def action_replay(self, player, *args, **kwargs):
		return await self.app.instance.command_manager.execute(player, '//replay')

	async def action_skip(self, player, *args, **kwargs):
		return await self.app.instance.command_manager.execute(player, '//skip')

	async def action_settings(self, player, *args, **kwargs):
		return await self.app.instance.command_manager.execute(player, '//settings')

	async def action_modesettings(self, player, *args, **kwargs):
		return await self.app.instance.command_manager.execute(player, '//modesettings')

	async def action_player_list(self, player, *args, **kwargs):
		return await self.app.instance.command_manager.execute(player, '//players')

	async def action_restart(self, player, *args, **kwargs):
		return await self.app.instance.command_manager.execute(player, '//restart')



"""
The UI Properties will be set and hold in the class definition bellow.
"""
import json
import logging
import xmltodict as xd

from pyplanet.core.signals import pyplanet_start_after
from pyplanet.core.ui.exceptions import UIPropertyDoesNotExist
from pyplanet.utils.functional import empty

logger = logging.getLogger(__name__)


class UIProperties:  # pragma: no cover
	"""
	Set the custom Script UI Properties.

	.. tip::

		Look at the possible UI Properties right here:

		- Trackmania: https://github.com/maniaplanet/script-xmlrpc/blob/master/XmlRpcListing.md#trackmaniauisetproperties
		- Shootmania: https://github.com/maniaplanet/script-xmlrpc/blob/master/XmlRpcListing.md#shootmaniauisetproperties
		- TM2020: Undocumented by Nadeo, check in-game tool OpenPlanet and discover what properties exist and what it means.

	Access this class with:

	.. code-block:: python

		self.instance.ui_manager.properties
	"""
	specials = ('_properties', '_instance')

	def __init__(self, instance):
		"""
		:param instance: Instance
		:type instance: pyplanet.core.instance.Instance
		"""
		self._instance = instance
		self._raw = None
		self._properties = dict()
		self._update_properties = dict()

	@property
	def properties(self):
		if self._instance.game.game in ['tm', 'sm']:
			if 'ui_properties' in self._properties:
				return self._properties['ui_properties']
		else:
			if self._properties:
				return self._properties
		return False

	async def on_start(self):
		await self.reset()
		await self.refresh_properties()
		self._instance.signals.listen(pyplanet_start_after, self.send_properties)

	async def reset(self):
		"""
		Reset the UI Properties to the default ManiaPlanet ones.
		:return:
		"""
		if self._instance.game.game == 'tm':
			method = 'Trackmania.UI.ResetProperties'
		elif self._instance.game.game == 'sm':
			method = 'Shootmania.UI.ResetProperties'
		else:
			method = 'Common.UIModules.ResetProperties'
		try:
			logger.debug('Resetting UIProperties...')
			await self._instance.gbx.script(method, response_id=False)
		except Exception as e:
			logger.warning('Unable to reset UIProperties: {}'.format(str(e)))

	async def refresh_properties(self):
		if self._instance.game.game == 'tm':
			method = 'Trackmania.UI.GetProperties'
		elif self._instance.game.game == 'sm':
			method = 'Shootmania.UI.GetProperties'
		else:
			method = 'Common.UIModules.GetProperties'
		try:
			if self._instance.game.game == 'tm' or self._instance.game.game == 'sm':
				self._raw = await self._instance.gbx(method, timeout=2)
				self._properties = xd.parse(self._raw['raw_1'])
			else:
				self._raw = await self._instance.gbx(method, timeout=2)
				self._properties = dict()
				for entry in self._raw['uimodules']:
					self._properties[entry['id']] = entry
		except Exception as e:
			self._properties = dict()
			self._raw = None

	def set_visibility(self, element: str, visible: bool):
		"""
		Set the visibility of the UI Property and don't complain about failing to set. Must be set at the start of the
		app(s).

		:param element: Element name, such as notices, map_info and chat.
						Full list: https://github.com/maniaplanet/script-xmlrpc/blob/master/XmlRpcListing.md#shootmaniauisetproperties
		:param visible: Boolean if the element should be visible.
		:return: Boolean, true if is set, false if failed to set.
		"""
		return self.set_attribute(element, 'visible', 'true' if visible else 'false')

	def get_visibility(self, element: str, default=empty):
		"""
		Set the visibility of the UI Property and don't complain about failing to set. Must be set at the start of the
		app(s).

		:param element: Element name, such as notices, map_info and chat.
						Full list: https://github.com/maniaplanet/script-xmlrpc/blob/master/XmlRpcListing.md#shootmaniauisetproperties
		:param default: The default value, or an exception if not given.
		:return: The boolean if it's visible or raise exception if not exists (or the default if default is given).
		"""
		value = self.get_attribute(element, 'visible', default)
		if isinstance(value, str):
			if value.lower() == 'true':
				return True
			elif value.lower() == 'false':
				return False
		return value

	def set_attribute(self, element: str, attribute: str, value):
		"""
		Set an attribute of an element and silent if it's not found. Useful to change positions but unsure if it will
		and still exists. Returns boolean if it's set successfully.

		:param element: Element name
		:param attribute: Attribute name
		:param value: New value of the attribute.
		:return: Boolean if it's set correctly.
		"""
		if not self._properties:
			return False
		if self._instance.game.game in ['tm', 'sm']:
			if element not in self.properties:
				return False
			if '@{}'.format(attribute) not in self.properties[element]:
				return False
			self.properties[element]['@{}'.format(attribute)] = value
		else:
			if element not in self.properties:
				return False

			self.properties[element][attribute] = value
			if element not in self._update_properties:
				self._update_properties[element] = dict(id=element)
			self._update_properties[element][attribute] = value
			self._update_properties[element]['{}_update'.format(attribute)] = True
		return True

	def get_attribute(self, element: str, attribute: str, default=empty):
		"""
		Get an attribute value of an element.

		:param element: Element name
		:param attribute: Attribute name
		:param default: Default if not found.
		:return: Boolean if it's set correctly.
		"""
		if not self._properties:
			if default is not empty:
				return default
			raise UIPropertyDoesNotExist('UI Properties are not present')
		if element not in self.properties:
			if default is not empty:
				return default
			raise UIPropertyDoesNotExist('UI Properties has no element with name \'{}\''.format(element))
		if self._instance.game.game in ['tm', 'sm'] and '@{}'.format(attribute) not in self.properties[element]:
			if default is not empty:
				return default
			raise UIPropertyDoesNotExist('UI Properties has no attribute with name \'{}\''.format(attribute))
		elif attribute not in self.properties[element]:
			if default is not empty:
				return default
			raise UIPropertyDoesNotExist('UI Properties has no attribute with name \'{}\''.format(attribute))

		if self._instance.game.game in ['tm', 'sm']:
			return self.properties[element]['@{}'.format(attribute)]
		return self.properties[element][attribute]

	async def send_properties(self, **kwargs):
		if not self._properties or (self._instance.game.game in ['tm', 'sm'] and 'ui_properties' not in self._properties):
			return

		# Decide the method to use.
		if self._instance.game.game == 'tm':
			method = 'Trackmania.UI.SetProperties'
		elif self._instance.game.game == 'sm':
			method = 'Shootmania.UI.SetProperties'
		else:
			method = 'Common.UIModules.SetProperties'

		if self._instance.game.game in ['tm', 'sm']:
			# Create XML document
			try:
				xml = xd.unparse(self._properties, full_document=False, short_empty_elements=True)
			except Exception as e:
				logger.warning('Can\'t convert UI Properties to XML document! Error: {}'.format(str(e)))
				return

			try:
				await self._instance.gbx(method, xml, encode_json=False, response_id=False)
			except Exception as e:
				logger.warning('Can\'t send UI Properties! Error: {}'.format(str(e)))
				return
		else:
			# Update TM2020 properties.
			try:
				await self._instance.gbx(
					method,
					json.dumps(dict(uimodules=list(self._update_properties.values()))),
					encode_json=False, response_id=False
				)
				logger.debug('UI Properties send and reloaded')
			except Exception as e:
				logger.warning('Can\'t send UI Properties! Error: {}'.format(str(e)))
				return



from pyplanet.utils.pip import Pip
from pyplanet.core.management import BaseCommand


class Command(BaseCommand):  # pragma: no cover
	help = 'Upgrade PyPlanet within the current PIP environment.'

	requires_system_checks = False
	requires_migrations_checks = False

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.pip = Pip()

	def add_arguments(self, parser):
		parser.add_argument(
			'--to-version', type=str, default=None,
			help='Upgrade to specific version given, leave empty to upgrade to latest'
		)

	def handle(self, *args, **options):
		version = options.get('to_version', None)
		print(options)

		if self.pip.is_supported:
			print('PIP: Found pip command: {}'.format(self.pip.command))
			print('PIP: ==> Supported!')
		else:
			print('PIP: ==>! Unsupported! Please manually upgrade your installation.')
			return

		print('')
		print('===================================================================================')
		print('! WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING !')
		print('The automatic update system can be unstable and can result in a broken installation')
		print('When the installation of PyPlanet is broken you have to use the manuals on the site')
		print('http://pypla.net')
		print('===================================================================================')

		user = input('Are you sure you want to upgrade PyPlanet to version: \'{}\'? [y/N]: '.format(
			version or 'latest'
		))
		if not user or not (user.lower() == 'y' or user.lower() == 'yes'):
			print('Cancelled!')
			return

		code, out, err = self.pip.install('pyplanet', target_version=version)

		if code == 0:
			print('Upgrade complete!')
		else:
			print('Error from PIP:')
			print(err.decode())



"""
Contrib packages are stated as less stable as the core itself. This way we can still develop and make PyPlanet better
on the long term without breaking any core code.

In the contrib package you can find many useful packages, such as the command manager and parameter parser.
"""


class CoreContrib:
	async def on_start(self):
		pass



from pyplanet.contrib import CoreContrib
from pyplanet.contrib.chat.query import ChatQuery


class ChatManager(CoreContrib):
	"""
	The Chat manager is available with: ``instance.chat`` shortcut.
	"""
	def __init__(self, instance):
		"""
		Initiate, should only be done from the core instance.
		
		:param instance: Instance.
		:type instance: pyplanet.core.instance.Instance
		"""
		self.instance = instance

	def __call__(self, *args, **kwargs):
		if len(args) <= 0:
			return
		query = self.prepare(args[0], raw=kwargs.get('raw', False))

		if len(args) > 1:
			query.to_players(args[1:])

		return query

	def prepare(self, message=None, raw=False):
		"""
		Prepare a Chat Query by returning a Chat Query object.
		
		:param message: Messsage predefined or build later.
		:param raw: Don't append prefixes or add any automatic message parts.
		:return: Query instance
		:rtype: pyplanet.contrib.chat.query.ChatQuery
		"""
		return ChatQuery(self, message, auto_prefix=not raw)

	def prepare_raw(self, message=None):
		"""
		Prepare raw message query without prefixes!
		
		:param message: Predefined message.
		:return: Query instance
		:rtype: pyplanet.contrib.chat.query.ChatQuery
		"""
		return self.prepare(message, True)

	async def execute(self, *queries):  # pragma: no cover
		"""
		Execute and send one or multiple chat messages (prepared queries or raw strings) with a multicall.
		
		:param queries: One or more query instances or one or multiple strings that gets send as global messages.
		:return: The results of the multicall.
		"""
		return await self.instance.gbx.multicall(
			*[
				q.gbx_query if isinstance(q, ChatQuery)
				else self.prepare_raw(str(q)).gbx_query

				for q in queries
			]
		)
